"""Some common rewrite functions to be shared between backends."""

from __future__ import annotations

from collections import defaultdict

import toolz

import xorq.vendor.ibis.expr.operations as ops
from xorq.common.exceptions import ExpressionError, XorqInputError
from xorq.vendor.ibis.common.collections import FrozenDict  # noqa: TCH001
from xorq.vendor.ibis.common.deferred import Item, _, deferred, var
from xorq.vendor.ibis.common.graph import Node as Traversable
from xorq.vendor.ibis.common.graph import traverse
from xorq.vendor.ibis.common.grounds import Concrete
from xorq.vendor.ibis.common.patterns import Check, pattern, replace
from xorq.vendor.ibis.common.typing import VarTuple  # noqa: TCH001
from xorq.vendor.ibis.util import Namespace, promote_list


p = Namespace(pattern, module=ops)
d = Namespace(deferred, module=ops)


x = var("x")
y = var("y")
name = var("name")


class DerefMap(Concrete, Traversable):
    """Trace and replace fields from earlier relations in the hierarchy.

    In order to provide a nice user experience, we need to allow expressions
    from earlier relations in the hierarchy. Consider the following example:

    t = ibis.table([('a', 'int64'), ('b', 'string')], name='t')
    t1 = t.select([t.a, t.b])
    t2 = t1.filter(t.a > 0)  # note that not t1.a is referenced here
    t3 = t2.select(t.a)  # note that not t2.a is referenced here

    However the relational operations in the IR are strictly enforcing that
    the expressions are referencing the immediate parent only. So we need to
    track fields upwards the hierarchy to replace `t.a` with `t1.a` and `t2.a`
    in the example above. This is called dereferencing.

    Whether we can treat or not a field of a relation semantically equivalent
    with a field of an earlier relation in the hierarchy depends on the
    `.values` mapping of the relation. Leaf relations, like `t` in the example
    above, have an empty `.values` mapping, so we cannot dereference fields
    from them. On the other hand a projection, like `t1` in the example above,
    has a `.values` mapping like `{'a': t.a, 'b': t.b}`, so we can deduce that
    `t1.a` is semantically equivalent with `t.a` and so on.
    """

    """The relations we want the values to point to."""
    rels: VarTuple[ops.Relation]

    """Substitution mapping from values of earlier relations to the fields of `rels`."""
    subs: FrozenDict[ops.Value, ops.Field]

    """Ambiguous field references."""
    ambigs: FrozenDict[ops.Value, VarTuple[ops.Value]]

    @classmethod
    def from_targets(cls, rels, extra=None):
        """Create a dereference map from a list of target relations.

        Usually a single relation is passed except for joins where multiple
        relations are involved.

        Parameters
        ----------
        rels : list of ops.Relation
            The target relations to dereference to.
        extra : dict, optional
            Extra substitutions to be added to the dereference map.

        Returns
        -------
        DerefMap
        """
        rels = promote_list(rels)
        mapping = defaultdict(dict)
        for rel in rels:
            for field in rel.fields.values():
                for value, distance in cls.backtrack(field):
                    mapping[value][field] = distance

        subs, ambigs = {}, {}
        for from_, to in mapping.items():
            mindist = min(to.values())
            minkeys = [k for k, v in to.items() if v == mindist]
            # if all the closest fields are from the same relation, then we
            # can safely substitute them and we pick the first one arbitrarily
            if all(minkeys[0].relations == k.relations for k in minkeys):
                subs[from_] = minkeys[0]
            else:
                ambigs[from_] = minkeys

        if extra is not None:
            subs.update(extra)

        return cls(rels, subs, ambigs)

    @classmethod
    def backtrack(cls, value):
        """Backtrack the field in the relation hierarchy.

        The field is traced back until no modification is made, so only follow
        ops.Field nodes not arbitrary values.

        Parameters
        ----------
        value : ops.Value
            The value to backtrack.

        Yields
        ------
        tuple[ops.Field, int]
            The value node and the distance from the original value.
        """
        distance = 0
        # track down the field in the hierarchy until no modification
        # is made so only follow ops.Field nodes not arbitrary values;
        while isinstance(value, ops.Field):
            yield value, distance
            value = value.rel.values.get(value.name)
            distance += 1
        if (
            value is not None
            and value.relations
            and not value.find(ops.Impure, filter=ops.Value)
        ):
            yield value, distance

    def dereference(self, value):
        """Dereference a value to the target relations.

        Also check for ambiguous field references. If a field reference is found
        which is marked as ambiguous, then raise an error.

        Parameters
        ----------
        value : ops.Value
            The value to dereference.

        Returns
        -------
        ops.Value
            The dereferenced value.
        """
        ambigs = value.find(lambda x: x in self.ambigs, filter=ops.Value)
        if ambigs:
            raise XorqInputError(
                f"Ambiguous field reference {ambigs!r} in expression {value!r}"
            )
        return value.replace(self.subs, filter=ops.Value)


def flatten_predicates(node):
    """Yield the expressions corresponding to the `And` nodes of a predicate.

    Examples
    --------
    >>> import ibis
    >>> t = ibis.table([("a", "int64"), ("b", "string")], name="t")
    >>> filt = (t.a == 1) & (t.b == "foo")
    >>> predicates = flatten_predicates(filt.op())
    >>> len(predicates)
    2
    >>> predicates[0].to_expr().name("left")
    r0 := UnboundTable: t
      a int64
      b string
    left: r0.a == 1
    >>> predicates[1].to_expr().name("right")
    r0 := UnboundTable: t
      a int64
      b string
    right: r0.b == 'foo'

    """

    def predicate(node):
        if isinstance(node, ops.And):
            # proceed and don't yield the node
            return True, None
        else:
            # halt and yield the node
            return False, node

    return list(traverse(predicate, node))


@replace(p.Field(p.JoinChain))
def peel_join_field(_):
    return _.rel.values[_.name]


@replace(p.ScalarParameter)
def replace_parameter(_, params, **kwargs):
    """Replace scalar parameters with their values."""
    return ops.Literal(value=params[_], dtype=_.dtype)


@replace(p.StringSlice)
def lower_stringslice(_, **kwargs):
    """Rewrite StringSlice in terms of Substring."""
    if _.end is None:
        return ops.Substring(_.arg, start=_.start)
    if _.start is None:
        return ops.Substring(_.arg, start=0, length=_.end)
    if (
        isinstance(_.start, ops.Literal)
        and isinstance(_.start.value, int)
        and isinstance(_.end, ops.Literal)
        and isinstance(_.end.value, int)
    ):
        # optimization for constant values
        length = _.end.value - _.start.value
    else:
        length = ops.Subtract(_.end, _.start)
    return ops.Substring(_.arg, start=_.start, length=length)


@replace(p.Analytic)
def wrap_analytic(_, **__):
    # Wrap analytic functions in a window function
    return ops.WindowFunction(_)


@replace(p.Reduction)
def project_wrap_reduction(_, rel):
    # Query all the tables that the reduction depends on
    if _.relations == {rel}:
        # The reduction is fully originating from the `rel`, so turn
        # it into a window function of `rel`
        return ops.WindowFunction(_, order_by=getattr(_, "order_by", ()))
    else:
        # 1. The reduction doesn't depend on any table, constructed from
        #    scalar values, so turn it into a scalar subquery.
        # 2. The reduction is originating from `rel` and other tables,
        #    so this is a correlated scalar subquery.
        # 3. The reduction is originating entirely from other tables,
        #    so this is an uncorrelated scalar subquery.
        return ops.ScalarSubquery(_.to_expr().as_table())


def rewrite_project_input(value, relation):
    # we need to detect reductions which are either turned into window functions
    # or scalar subqueries depending on whether they are originating from the
    # relation
    return value.replace(
        wrap_analytic | project_wrap_reduction,
        filter=p.Value & ~p.WindowFunction,
        context={"rel": relation},
    )


ReductionLike = p.Reduction | p.Field(p.Aggregate(groups={}))


@replace(ReductionLike)
def filter_wrap_reduction(_):
    # Wrap reductions or fields referencing an aggregation without a group by -
    # which are scalar fields - in a scalar subquery. In the latter case we
    # use the reduction value from the aggregation.
    if isinstance(_, ops.Field):
        value = _.rel.values[_.name]
    else:
        value = _
    return ops.ScalarSubquery(value.to_expr().as_table())


def rewrite_filter_input(value):
    return value.replace(
        wrap_analytic | filter_wrap_reduction, filter=p.Value & ~p.WindowFunction
    )


@replace(p.Analytic | p.Reduction)
def window_wrap_reduction(_, window):
    # Wrap analytic and reduction functions in a window function. Used in the
    # value.over() API.
    return ops.WindowFunction(
        _,
        how=window.how,
        start=window.start,
        end=window.end,
        group_by=window.groupings,
        order_by=window.orderings,
    )


@replace(p.WindowFunction)
def window_merge_frames(_, window):
    # Merge window frames, used in the value.over() and groupby.select() APIs.
    if _.how != window.how:
        raise ExpressionError(
            f"Unable to merge {_.how} window with {window.how} window"
        )
    elif _.start and window.start and _.start != window.start:
        raise ExpressionError(
            "Unable to merge windows with conflicting `start` boundary"
        )
    elif _.end and window.end and _.end != window.end:
        raise ExpressionError("Unable to merge windows with conflicting `end` boundary")

    start = _.start or window.start
    end = _.end or window.end
    group_by = tuple(toolz.unique(_.group_by + window.groupings))

    order_keys = {}
    for sort_key in window.orderings + _.order_by:
        order_keys[sort_key.expr] = sort_key.ascending, sort_key.nulls_first

    order_by = (
        ops.SortKey(expr, ascending=ascending, nulls_first=nulls_first)
        for expr, (ascending, nulls_first) in order_keys.items()
    )
    return _.copy(start=start, end=end, group_by=group_by, order_by=order_by)


def rewrite_window_input(value, window):
    context = {"window": window}
    # if self is a reduction or analytic function, wrap it in a window function
    node = value.replace(
        window_wrap_reduction,
        filter=p.Value & ~p.WindowFunction,
        context=context,
    )
    # if self is already a window function, merge the existing window frame
    # with the requested window frame
    return node.replace(window_merge_frames, filter=p.Value, context=context)


# TODO(kszucs): schema comparison should be updated to not distinguish between
# different column order
@replace(p.Project(y @ p.Relation) & Check(_.schema == y.schema))
def complete_reprojection(_, y):
    # TODO(kszucs): this could be moved to the pattern itself but not sure how
    # to express it, especially in a shorter way then the following check
    for name in _.schema:
        if _.values[name] != ops.Field(y, name):
            return _
    return y


@replace(p.Project(y @ p.Project))
def subsequent_projects(_, y):
    rule = p.Field(y, name) >> Item(y.values, name)
    values = {k: v.replace(rule, filter=ops.Value) for k, v in _.values.items()}
    return ops.Project(y.parent, values)


@replace(p.Filter(y @ p.Filter))
def subsequent_filters(_, y):
    rule = p.Field(y, name) >> d.Field(y.parent, name)
    preds = tuple(v.replace(rule, filter=ops.Value) for v in _.predicates)
    return ops.Filter(y.parent, y.predicates + preds)


@replace(p.Filter(y @ p.Project))
def reorder_filter_project(_, y):
    rule = p.Field(y, name) >> Item(y.values, name)
    preds = tuple(v.replace(rule, filter=ops.Value) for v in _.predicates)

    inner = ops.Filter(y.parent, preds)
    rule = p.Field(y.parent, name) >> d.Field(inner, name)
    projs = {k: v.replace(rule, filter=ops.Value) for k, v in y.values.items()}

    return ops.Project(inner, projs)


def simplify(node):
    # TODO(kszucs): add a utility to the graph module to do rewrites in multiple
    # passes after each other
    node = node.replace(reorder_filter_project)
    node = node.replace(reorder_filter_project)
    node = node.replace(subsequent_projects | subsequent_filters)
    node = node.replace(complete_reprojection)
    return node



from __future__ import annotations

import decimal
from operator import invert, neg

import numpy as np
import pandas as pd
import pytest
import toolz
from pytest import param

import xorq.api as xo
import xorq.common.exceptions as com
import xorq.expr.datatypes as dt
from xorq.tests.util import assert_frame_equal, assert_series_equal
from xorq.vendor.ibis import _
from xorq.vendor.ibis.common.annotations import ValidationError


def test_null_literal(con):
    expr = xo.null()
    assert pd.isna(con.execute(expr))
    assert con.execute(expr.typeof()) == "Null"

    assert expr.type() == dt.null
    assert pd.isna(con.execute(expr.cast(str).upper()))


def test_boolean_literal(con):
    expr = xo.literal(False, type=dt.boolean)
    result = con.execute(expr)
    assert not result
    assert type(result) in (np.bool_, bool)
    assert con.execute(expr.typeof()) == "Boolean"


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        param(xo.null().fill_null(5), 5, id="na_fillna"),
        param(xo.literal(5).fill_null(10), 5, id="non_na_fillna"),
        param(xo.literal(5).nullif(5), None, id="nullif_null"),
        param(xo.literal(10).nullif(5), 10, id="nullif_not_null"),
    ],
)
def test_scalar_fillna_nullif(con, expr, expected):
    if expected is None:
        assert pd.isna(con.execute(expr))
    else:
        assert con.execute(expr) == expected


@pytest.mark.parametrize(
    ("col", "filt"),
    [
        param(
            "nan_col",
            _.nan_col.isnan(),
            id="nan_col",
        ),
        param(
            "none_col",
            _.none_col.isnull(),
            id="none_col",
        ),
    ],
)
def test_isna(alltypes, col, filt):
    table = alltypes.select(
        nan_col=xo.literal(np.nan), none_col=xo.null().cast("float64")
    )
    df = xo.execute(table)

    result = table[filt].execute().reset_index(drop=True)
    expected = df[df[col].isna()].reset_index(drop=True)

    assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    "value",
    [
        None,
    ],
)
def test_column_fillna(alltypes, value):
    table = alltypes.mutate(missing=xo.literal(value).cast("float64"))
    pd_table = xo.execute(table)

    res = table.mutate(missing=table.missing.fill_null(0.0)).execute()
    sol = pd_table.assign(missing=pd_table.missing.fillna(0.0))
    assert_frame_equal(res.reset_index(drop=True), sol.reset_index(drop=True))


@pytest.mark.parametrize(
    ("expr", "expected"),
    [
        param(xo.coalesce(5, None, 4), 5, id="generic"),
        param(xo.coalesce(xo.null(), 4, xo.null()), 4, id="null_start_end"),
        param(
            xo.coalesce(xo.null(), xo.null(), 3.14),
            3.14,
            id="non_null_last",
        ),
    ],
)
def test_coalesce(con, expr, expected):
    result = con.execute(expr.name("tmp"))

    if isinstance(result, decimal.Decimal):
        # in case of Impala the result is decimal
        # >>> decimal.Decimal('5.56') == 5.56
        # False
        assert result == decimal.Decimal(str(expected))
    else:
        assert result == pytest.approx(expected)


@pytest.mark.parametrize(
    ("column", "elements"),
    [
        ("int_col", [1, 2, 3]),
        ("int_col", (1, 2, 3)),
        ("string_col", ["1", "2", "3"]),
        ("string_col", ("1", "2", "3")),
        ("int_col", {1}),
        ("int_col", frozenset({1})),
    ],
)
def test_isin(alltypes, sorted_df, column, elements):
    sorted_alltypes = alltypes.order_by("id")
    expr = sorted_alltypes[
        "id", sorted_alltypes[column].isin(elements).name("tmp")
    ].order_by("id")
    result = expr.execute().tmp

    expected = sorted_df[column].isin(elements)
    assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ("column", "elements"),
    [
        ("int_col", [1, 2, 3]),
        ("int_col", (1, 2, 3)),
        ("string_col", ["1", "2", "3"]),
        ("string_col", ("1", "2", "3")),
        ("int_col", {1}),
        ("int_col", frozenset({1})),
    ],
)
def test_notin(alltypes, sorted_df, column, elements):
    sorted_alltypes = alltypes.order_by("id")
    expr = sorted_alltypes[
        "id", sorted_alltypes[column].notin(elements).name("tmp")
    ].order_by("id")
    result = expr.execute().tmp

    expected = ~sorted_df[column].isin(elements)
    assert_series_equal(result, expected)


@pytest.mark.parametrize(
    ("predicate_fn", "expected_fn"),
    [
        param(lambda t: t["bool_col"], lambda df: df["bool_col"], id="no_op"),
        param(lambda t: ~t["bool_col"], lambda df: ~df["bool_col"], id="negate"),
        param(
            lambda t: t.bool_col & t.bool_col,
            lambda df: df.bool_col & df.bool_col,
            id="and",
        ),
        param(
            lambda t: t.bool_col | t.bool_col,
            lambda df: df.bool_col | df.bool_col,
            id="or",
        ),
        param(
            lambda t: t.bool_col ^ t.bool_col,
            lambda df: df.bool_col ^ df.bool_col,
            id="xor",
        ),
    ],
)
def test_filter(alltypes, sorted_df, predicate_fn, expected_fn):
    sorted_alltypes = alltypes.order_by("id")
    table = sorted_alltypes[predicate_fn(sorted_alltypes)].order_by("id")
    result = table.execute()
    expected = sorted_df[expected_fn(sorted_df)]

    assert_frame_equal(result, expected)


def test_case_where(alltypes, alltypes_df):
    table = alltypes
    table = table.mutate(
        new_col=(
            xo.cases(
                (table["int_col"] == 1, 20),
                (table["int_col"] == 0, 10),
                else_=0,
            ).cast("int64")
        )
    )

    result = table.execute()

    expected = alltypes_df.copy()
    mask_0 = expected["int_col"] == 1
    mask_1 = expected["int_col"] == 0

    expected["new_col"] = 0
    expected.loc[mask_0, "new_col"] = 20
    expected.loc[mask_1, "new_col"] = 10

    assert_frame_equal(result, expected)


def test_table_fill_null_invalid(alltypes):
    with pytest.raises(
        com.XorqTypeError, match=r"Column 'invalid_col' is not found in table"
    ):
        alltypes.fill_null({"invalid_col": 0.0})

    with pytest.raises(
        com.XorqTypeError, match="Cannot fill_null on column 'string_col' of type.*"
    ):
        alltypes[["int_col", "string_col"]].fill_null(0)

    with pytest.raises(
        com.XorqTypeError, match="Cannot fill_null on column 'int_col' of type.*"
    ):
        alltypes.fill_null({"int_col": "oops"})


@pytest.mark.parametrize(
    "replacements",
    [
        param({"int_col": 20}, id="int"),
        param(
            {"double_col": -1, "string_col": "missing"},
            id="double-int-str",
        ),
        param(
            {"double_col": -1.5, "string_col": "missing"},
            id="double-str",
        ),
    ],
)
def test_table_fillna_mapping(alltypes, replacements):
    table = alltypes.mutate(
        int_col=alltypes.int_col.nullif(1),
        double_col=alltypes.double_col.nullif(3.0),
        string_col=alltypes.string_col.nullif("2"),
    ).select("id", "int_col", "double_col", "string_col")
    pd_table = table.execute()

    result = table.fill_null(replacements).execute().reset_index(drop=True)
    expected = pd_table.fillna(replacements).reset_index(drop=True)

    assert_frame_equal(result, expected, check_dtype=False)


def test_table_fillna_scalar(alltypes):
    table = alltypes.mutate(
        int_col=alltypes.int_col.nullif(1),
        double_col=alltypes.double_col.nullif(3.0),
        string_col=alltypes.string_col.nullif("2"),
    ).select("id", "int_col", "double_col", "string_col")
    pd_table = table.execute()

    res = table[["int_col", "double_col"]].fill_null(0).execute().reset_index(drop=True)
    sol = pd_table[["int_col", "double_col"]].fillna(0).reset_index(drop=True)
    assert_frame_equal(res, sol, check_dtype=False)

    res = table[["string_col"]].fill_null("missing").execute().reset_index(drop=True)
    sol = pd_table[["string_col"]].fillna("missing").reset_index(drop=True)
    assert_frame_equal(res, sol, check_dtype=False)


def test_mutate_rename(alltypes):
    table = alltypes.select(["bool_col", "string_col"])
    table = table.mutate(dupe_col=table["bool_col"])
    result = table.execute()
    # check_dtype is False here because there are dtype diffs between
    # Pyspark and Pandas on Java 8 - filling the 'none_col' with an int
    # results in float in Pyspark, and int in Pandas. This diff does
    # not exist in Java 11.
    assert list(result.columns) == ["bool_col", "string_col", "dupe_col"]


def test_drop_null_invalid(alltypes):
    with pytest.raises(
        com.XorqTypeError, match=r"Column 'invalid_col' is not found in table"
    ):
        alltypes.drop_null(subset=["invalid_col"])

    with pytest.raises(ValidationError):
        alltypes.drop_null(how="invalid")


@pytest.mark.parametrize("how", ["any", "all"])
@pytest.mark.parametrize(
    "subset", [None, [], "col_1", ["col_1", "col_2"], ["col_1", "col_3"]]
)
def test_dropna_table(alltypes, how, subset):
    is_two = alltypes.int_col == 2
    is_four = alltypes.int_col == 4

    table = alltypes.mutate(
        col_1=is_two.ifelse(xo.null(), alltypes.float_col),
        col_2=is_four.ifelse(xo.null(), alltypes.float_col),
        col_3=(is_two | is_four).ifelse(xo.null(), alltypes.float_col),
    ).select("col_1", "col_2", "col_3")

    table_pandas = xo.execute(table)
    result = table.drop_null(subset, how).execute().reset_index(drop=True)
    expected = table_pandas.dropna(how=how, subset=subset).reset_index(drop=True)

    assert_frame_equal(result, expected)


def test_select_sort_sort(alltypes):
    query = alltypes[alltypes.year, alltypes.bool_col]
    query = query.order_by(query.year).order_by(query.bool_col)


@pytest.mark.parametrize(
    "key, df_kwargs",
    [
        param("id", {"by": "id"}),
        param(_.id, {"by": "id"}),
        param(lambda _: _.id, {"by": "id"}),
        param(
            xo.desc("id"),
            {"by": "id", "ascending": False},
        ),
        param(
            ["id", "int_col"],
            {"by": ["id", "int_col"]},
        ),
        param(
            ["id", xo.desc("int_col")],
            {"by": ["id", "int_col"], "ascending": [True, False]},
        ),
    ],
)
def test_order_by(alltypes, alltypes_df, key, df_kwargs):
    result = alltypes.filter(_.id < 100).order_by(key).execute()
    expected = alltypes_df.loc[alltypes_df.id < 100].sort_values(**df_kwargs)
    assert_frame_equal(result, expected)


def test_order_by_random(alltypes):
    expr = alltypes.filter(_.id < 100).order_by(xo.random()).limit(5)
    r1 = xo.execute(expr)
    r2 = xo.execute(expr)
    assert len(r1) == 5
    assert len(r2) == 5
    # Ensure that multiple executions returns different results
    assert not r1.equals(r2)


@pytest.mark.parametrize(
    ("ibis_op", "pandas_op"),
    [
        param(
            _.string_col.isin([]),
            lambda df: df.string_col.isin([]),
            id="isin",
        ),
        param(
            _.string_col.notin([]),
            lambda df: ~df.string_col.isin([]),
            id="notin",
        ),
        param(
            (_.string_col.length() * 1).isin([1]),
            lambda df: (df.string_col.str.len() * 1).isin([1]),
            id="isin_non_empty",
        ),
        param(
            (_.string_col.length() * 1).notin([1]),
            lambda df: ~(df.string_col.str.len() * 1).isin([1]),
            id="notin_non_empty",
        ),
    ],
)
def test_isin_notin(alltypes, alltypes_df, ibis_op, pandas_op):
    expr = alltypes[ibis_op]
    expected = (
        alltypes_df.loc[pandas_op(alltypes_df)]
        .sort_values(["id"])
        .reset_index(drop=True)
    )
    result = expr.execute().sort_values(["id"]).reset_index(drop=True)
    assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    ("expr", "expected", "op"),
    [
        param(True, True, toolz.identity, id="true_noop"),
        param(False, False, toolz.identity, id="false_noop"),
        param(True, False, invert, id="true_invert"),
        param(False, True, invert, id="false_invert"),
        param(True, False, neg, id="true_negate"),
        param(False, True, neg, id="false_negate"),
    ],
)
def test_logical_negation_literal(con, expr, expected, op):
    assert con.execute(op(xo.literal(expr)).name("tmp")) == expected


@pytest.mark.parametrize(
    "op",
    [
        toolz.identity,
        invert,
        neg,
    ],
)
def test_logical_negation_column(alltypes, alltypes_df, op):
    result = op(alltypes["bool_col"]).name("tmp").execute()
    expected = op(alltypes_df["bool_col"])
    assert_series_equal(result, expected, check_names=False)


def test_ifelse_select(alltypes, alltypes_df):
    table = alltypes
    table = table.select(
        [
            "int_col",
            (xo.ifelse(table["int_col"] == 0, 42, -1).cast("int64").name("where_col")),
        ]
    )

    result = table.execute()

    expected = alltypes_df.loc[:, ["int_col"]].copy()

    expected["where_col"] = -1
    expected.loc[expected["int_col"] == 0, "where_col"] = 42

    assert_frame_equal(result, expected)


def test_ifelse_column(alltypes, alltypes_df):
    expr = xo.ifelse(alltypes["int_col"] == 0, 42, -1).cast("int64").name("where_col")
    result = xo.execute(expr)

    expected = pd.Series(
        np.where(alltypes_df.int_col == 0, 42, -1),
        name="where_col",
        dtype="int64",
    )

    assert_series_equal(result, expected)


def test_select_filter(alltypes, alltypes_df):
    t = alltypes

    expr = t.select("int_col", "string_col").filter(t.string_col == "4")
    result = expr.execute()

    expected = alltypes_df.loc[
        alltypes_df.string_col == "4", ["int_col", "string_col"]
    ].reset_index(drop=True)
    assert_frame_equal(result, expected)


def test_select_filter_select(alltypes, alltypes_df):
    t = alltypes
    expr = t.select("int_col", "string_col").filter(t.string_col == "4").int_col
    result = expr.execute().rename("int_col")

    expected = alltypes_df.loc[alltypes_df.string_col == "4", "int_col"].reset_index(
        drop=True
    )
    assert_series_equal(result, expected)


def test_interactive(alltypes, monkeypatch):
    monkeypatch.setattr(xo.options, "interactive", True)

    expr = alltypes.mutate(
        str_col=_.string_col.replace("1", "").nullif("2"),
        date_col=_.timestamp_col.date(),
        delta_col=lambda t: xo.now() - t.timestamp_col,
    )

    repr(expr)


def test_correlated_subquery(alltypes):
    expr = alltypes[_.double_col > _.view().double_col]
    assert expr.compile() is not None


def test_uncorrelated_subquery(batting, batting_df):
    subset_batting = batting[batting.yearID <= 2000]
    expr = batting[_.yearID == subset_batting.yearID.max()]["playerID", "yearID"]
    result = expr.execute()

    expected = batting_df[batting_df.yearID == 2000][["playerID", "yearID"]]
    assert_frame_equal(result, expected)


def test_int_column(alltypes):
    expr = alltypes.mutate(x=xo.literal(1)).x
    result = expr.execute()
    assert expr.type() == dt.int8
    assert result.dtype == np.int8


def test_int_scalar(alltypes):
    expr = alltypes.smallint_col.min()
    result = expr.execute()
    assert expr.type() == dt.int16
    assert result.dtype == np.int16


@pytest.mark.parametrize(
    "dtype",
    [
        "bool",
        "bytes",
        "str",
        "int",
        "float",
        "int8",
        "int16",
        "int32",
        "int64",
        "float32",
        "float64",
        "timestamp",
        "date",
        "time",
    ],
)
def test_literal_na(con, dtype):
    expr = xo.literal(None, type=dtype)
    result = con.execute(expr)
    assert pd.isna(result)


def test_memtable_bool_column(con):
    t = xo.memtable({"a": [True, False, True]})
    assert_series_equal(con.execute(t.a), pd.Series([True, False, True], name="a"))


def test_memtable_construct():
    pa = pytest.importorskip("pyarrow")

    pa_t = pa.Table.from_pydict(
        {
            "a": list("abc"),
            "b": [1, 2, 3],
            "c": [1.0, 2.0, 3.0],
            "d": [None, "b", None],
        }
    )
    t = xo.memtable(pa_t)
    assert_frame_equal(xo.execute(t).fillna(pd.NA), pa_t.to_pandas().fillna(pd.NA))


def test_pivot_wider(diamonds):
    expr = (
        diamonds.group_by(["cut", "color"])
        .agg(carat=_.carat.mean())
        .pivot_wider(
            names_from="cut", values_from="carat", names_sort=True, values_agg="mean"
        )
    )
    df = expr.execute()
    assert set(df.columns) == {"color"} | set(
        diamonds[["cut"]].distinct().cut.execute()
    )
    assert len(df) == diamonds.color.nunique().execute()


@pytest.mark.parametrize(
    ("slc", "expected_count_fn"),
    [
        ###################
        ### NONE/ZERO start
        # no stop
        param(slice(None, 0), lambda _: 0, id="[:0]"),
        param(slice(None, None), lambda t: t.count().to_pandas(), id="[:]"),
        param(slice(0, 0), lambda _: 0, id="[0:0]"),
        param(slice(0, None), lambda t: t.count().to_pandas(), id="[0:]"),
        # positive stop
        param(slice(None, 2), lambda _: 2, id="[:2]"),
        param(slice(0, 2), lambda _: 2, id="[0:2]"),
        ##################
        ### NEGATIVE start
        # zero stop
        param(slice(-3, 0), lambda _: 0, id="[-3:0]"),
        # negative stop
        param(slice(-3, -3), lambda _: 0, id="[-3:-3]"),
        param(slice(-3, -4), lambda _: 0, id="[-3:-4]"),
        param(slice(-3, -5), lambda _: 0, id="[-3:-5]"),
        ##################
        ### POSITIVE start
        # no stop
        param(slice(3, 0), lambda _: 0, id="[3:0]"),
        param(
            slice(3, None),
            lambda t: t.count().to_pandas() - 3,
            id="[3:]",
        ),
        # positive stop
        param(slice(3, 2), lambda _: 0, id="[3:2]"),
        param(
            slice(3, 4),
            lambda _: 1,
            id="[3:4]",
        ),
    ],
)
def test_static_table_slice(slc, expected_count_fn, alltypes):
    t = alltypes

    rows = t[slc]
    count = rows.count().to_pandas()

    expected_count = expected_count_fn(t)
    assert count == expected_count


def test_sample(alltypes):
    t = alltypes.filter(_.int_col >= 2)

    total_rows = t.count().execute()
    empty = t.limit(1).execute().iloc[:0]

    df = t.sample(0.1, method="row").execute()
    assert len(df) <= total_rows
    assert_frame_equal(empty, df.iloc[:0])

    df = t.sample(0.1, method="block").execute()
    assert len(df) <= total_rows
    assert_frame_equal(empty, df.iloc[:0])


def test_sample_memtable(con):
    df = pd.DataFrame({"x": [1, 2, 3, 4]})
    res = con.execute(xo.memtable(df).sample(0.5))
    assert len(res) <= 4
    assert_frame_equal(res.iloc[:0], df.iloc[:0])


def test_hexdigest(alltypes):
    h1 = alltypes.order_by("id").string_col.hexdigest().execute(limit=10)
    df = alltypes.order_by("id").execute(limit=10)

    import hashlib

    def hash_256(col):
        return hashlib.sha256(col.encode()).hexdigest()

    h2 = df["string_col"].apply(hash_256).rename("HexDigest(string_col)")

    assert_series_equal(h1, h2)


def test_typeof(con):
    # Other tests also use the typeof operation, but only this test has this operation required.
    expr = xo.literal(1).typeof()
    result = con.execute(expr)

    assert result is not None



from __future__ import annotations

import contextlib
import functools
import inspect
import json
import typing
from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pyarrow as pa
import pyarrow_hotfix  # noqa: F401
import sqlglot as sg
import sqlglot.expressions as sge

import xorq
import xorq.common.exceptions as com
import xorq.expr.datatypes as dt
import xorq.internal as df
import xorq.vendor.ibis.expr.operations as ops
import xorq.vendor.ibis.expr.schema as sch
import xorq.vendor.ibis.expr.types as ir
from xorq.backends.xorq.datafusion.compiler import compiler
from xorq.backends.xorq.datafusion.provider import IbisTableProvider
from xorq.common.utils import classproperty
from xorq.common.utils.aws_utils import (
    make_s3_connection,
)
from xorq.expr import Expr
from xorq.expr.pyaggregator import PyAggregator, make_struct_type
from xorq.expr.udf import ExprScalarUDF
from xorq.internal import (
    DataFrame,
    SessionConfig,
    SessionContext,
    Table,
    WindowEvaluator,
    udwf,
)
from xorq.vendor import ibis
from xorq.vendor.ibis.backends import (
    CanCreateCatalog,
    CanCreateDatabase,
    CanCreateSchema,
    NoUrl,
)
from xorq.vendor.ibis.backends.sql import SQLBackend
from xorq.vendor.ibis.backends.sql.compilers.base import C
from xorq.vendor.ibis.common.dispatch import lazy_singledispatch
from xorq.vendor.ibis.expr.operations.udf import InputType
from xorq.vendor.ibis.formats.pyarrow import (
    PyArrowType,
)
from xorq.vendor.ibis.util import gen_name, normalize_filename, normalize_filenames


if TYPE_CHECKING:
    import pandas as pd


def _compile_pyarrow_udwf(udwf_node):
    def make_datafusion_udwf(
        input_types,
        return_type,
        name,
        evaluate=None,
        evaluate_all=None,
        evaluate_all_with_rank=None,
        supports_bounded_execution=False,
        uses_window_frame=False,
        include_rank=False,
        volatility="immutable",
        **kwargs,
    ):
        def return_value(value):
            def f(_):
                return value

            return f

        kwds = {
            "evaluate": evaluate,
            "evaluate_all": evaluate_all,
            "evaluate_all_with_rank": evaluate_all_with_rank,
            "supports_bounded_execution": return_value(supports_bounded_execution),
            "uses_window_frame": return_value(uses_window_frame),
            "include_rank": return_value(include_rank),
            **kwargs,
        }
        mytyp = type(
            name,
            (WindowEvaluator,),
            kwds,
        )
        my_udwf = udwf(
            mytyp(),
            input_types,
            return_type,
            volatility=str(volatility),
            # datafusion normalizes to lower case and ibis doesn't quote
            name=name.lower(),
        )
        return my_udwf

    config = udwf_node.__config__ | {
        "input_types": tuple(
            dt.to_pyarrow() for dt in udwf_node.__config__["input_types"]
        ),
        "return_type": udwf_node.__config__["return_type"].to_pyarrow(),
    }

    my_udwf = make_datafusion_udwf(**config)
    return my_udwf


def _compile_pyarrow_udaf(udaf_node):
    func = udaf_node.__func__
    name = type(udaf_node).__name__
    return_type = PyArrowType.from_ibis(udaf_node.dtype)
    parameters = (
        (name, PyArrowType.from_ibis(param.annotation.pattern.dtype))
        for name, param in udaf_node.__signature__.parameters.items()
        if name != "where"
    )
    names, input_types = map(list, zip(*parameters))  # noqa
    struct_type = make_struct_type(names, input_types)

    class MyAggregator(PyAggregator):
        @classproperty
        def struct_type(cls):
            return struct_type

        def py_evaluate(self):
            struct_array = self.pystate()
            args = (struct_array.field(field_name) for field_name in self.names)
            return func(*args)

        @classproperty
        def return_type(cls):
            return return_type

        @classproperty
        def name(cls):
            return name

    return df.udaf(
        accum=MyAggregator,
        input_type=input_types,
        return_type=return_type,
        state_type=[MyAggregator.state_type],
        volatility=MyAggregator.volatility,
        name=name,
    )


def _inspect_xgboost_model_from_json(json_file_path):
    with open(json_file_path, "r") as file:
        model_data = json.load(file)

    learner_data = model_data["learner"]
    model_attributes = learner_data["attributes"]
    feature_names = learner_data["feature_names"]
    feature_types = learner_data["feature_types"]
    gbtree_model_param = learner_data["gradient_booster"]["model"]["gbtree_model_param"]
    num_trees = gbtree_model_param["num_trees"]

    metadata = {
        "model_attributes": model_attributes,
        "feature_names": feature_names,
        "feature_types": feature_types,
        "gbtree_model_param": gbtree_model_param,
        "number_of_trees": num_trees,
    }

    return metadata


def _fields_to_parameters(fields):
    parameters = []
    for name, arg in fields.items():
        param = inspect.Parameter(
            name, inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=arg.typehint
        )
        parameters.append(param)
    return parameters


def translate_sort(exprs: list[ir.Expr]):
    result = []
    for expr in exprs:
        node = expr.op()
        if not isinstance(node, ops.SortKey):
            raise ValueError(f"Expected SortKey, got {type(node)}")

        column_identifier = str(sg.to_identifier(node.name, quoted=True))
        result.append(
            Expr.column(column_identifier).sort(node.ascending, node.nulls_first)
        )

    return result


class Backend(SQLBackend, CanCreateCatalog, CanCreateDatabase, CanCreateSchema, NoUrl):
    name = "datafusion"
    supports_in_memory_tables = True
    supports_arrays = True
    compiler = compiler

    @property
    def version(self):
        return xorq.__version__

    def do_connect(self, config: SessionConfig | None = None) -> None:
        if config is None:
            config = SessionConfig()
        config = config.with_information_schema(True).set(
            "datafusion.sql_parser.dialect", "PostgreSQL"
        )

        self.con = SessionContext(config=config)

        self._register_builtin_udfs()

    def disconnect(self) -> None:
        pass

    @contextlib.contextmanager
    def _safe_raw_sql(self, sql: sge.Statement) -> Any:
        yield self.raw_sql(sql).collect()

    def _get_schema_using_query(self, query: str) -> sch.Schema:
        name = gen_name("datafusion_metadata_view")
        table = sg.table(name, quoted=self.compiler.quoted)
        src = sge.Create(
            this=table,
            kind="VIEW",
            expression=sg.parse_one(query, read="datafusion"),
        )

        with self._safe_raw_sql(src):
            pass

        try:
            result = (
                self.raw_sql(f"DESCRIBE {table.sql(self.dialect)}")
                .to_arrow_table()
                .to_pydict()
            )
        finally:
            self.drop_view(name)
        return sch.Schema(
            {
                name: self.compiler.type_mapper.from_string(
                    type_string, nullable=is_nullable == "YES"
                )
                for name, type_string, is_nullable in zip(
                    result["column_name"], result["data_type"], result["is_nullable"]
                )
            }
        )

    def _register_builtin_udfs(self):
        from xorq.backends.xorq.datafusion import udfs

        for name, func in inspect.getmembers(
            udfs,
            predicate=lambda m: (
                callable(m)
                and not m.__name__.startswith("_")
                and m.__module__ == udfs.__name__
            ),
        ):
            annotations = typing.get_type_hints(func)
            argnames = list(inspect.signature(func).parameters.keys())
            input_types = [
                PyArrowType.from_ibis(dt.dtype(annotations.get(arg_name)))
                for arg_name in argnames
            ]
            return_type = PyArrowType.from_ibis(dt.dtype(annotations["return"]))
            udf = df.udf(
                func,
                input_types=input_types,
                return_type=return_type,
                volatility="immutable",
                name=name,
            )
            self.con.register_udf(udf)

    def _register_udfs(self, expr: ir.Expr) -> None:
        for udf_node in expr.op().find(ops.ScalarUDF):
            if udf_node.__input_type__ == InputType.PYARROW:
                if isinstance(udf_node, ExprScalarUDF):
                    udf = self._compile_pyarrow_expr_udf(udf_node)
                else:
                    udf = self._compile_pyarrow_udf(udf_node)
                self.con.register_udf(udf)

        for udf_node in expr.op().find(ops.ElementWiseVectorizedUDF):
            udf = self._compile_elementwise_udf(udf_node)
            self.con.register_udf(udf)

        for agg_node in expr.op().find(ops.AggUDF):
            if agg_node.__input_type__ == InputType.PYARROW:
                if {"evaluate", "evaluate_all"}.intersection(agg_node.__config__):
                    udwf = _compile_pyarrow_udwf(agg_node)
                    self.con.register_udwf(udwf)
                else:
                    udaf = _compile_pyarrow_udaf(agg_node)
                    self.con.register_udaf(udaf)

    def _compile_pyarrow_expr_udf(self, udf_node):
        def extract_computed_arg(expr):
            # user can do Scalar.to_table() if they want to cache it
            import pandas as pd

            value = expr.execute()
            if isinstance(value, pd.DataFrame):
                if value.shape != (1, 1):
                    raise ValueError
                ((value,),) = value.values
            value = udf_node.post_process_fn(value)
            return value

        computed_arg = extract_computed_arg(udf_node.computed_kwargs_expr)
        return df.udf(
            functools.partial(udf_node.__func__, computed_arg=computed_arg),
            input_types=[PyArrowType.from_ibis(arg.dtype) for arg in udf_node.args],
            return_type=PyArrowType.from_ibis(udf_node.dtype),
            volatility=getattr(udf_node, "__config__", {}).get(
                "volatility", "volatile"
            ),
            name=udf_node.__func_name__,
        )

    def _compile_pyarrow_udf(self, udf_node):
        return df.udf(
            udf_node.__func__,
            input_types=[PyArrowType.from_ibis(arg.dtype) for arg in udf_node.args],
            return_type=PyArrowType.from_ibis(udf_node.dtype),
            volatility=getattr(udf_node, "__config__", {}).get(
                "volatility", "volatile"
            ),
            name=udf_node.__func_name__,
        )

    def _compile_elementwise_udf(self, udf_node):
        return df.udf(
            udf_node.func,
            input_types=list(map(PyArrowType.from_ibis, udf_node.input_type)),
            return_type=PyArrowType.from_ibis(udf_node.return_type),
            volatility="volatile",
            name=udf_node.func.__name__,
        )

    def raw_sql(self, query: str | sge.Expression) -> Any:
        """Execute a SQL string `query` against the database.

        Parameters
        ----------
        query
            Raw SQL string
        """
        with contextlib.suppress(AttributeError):
            query = query.sql(dialect=self.dialect, pretty=True)
        self._log(query)
        return self.con.sql(query)

    @property
    def current_catalog(self) -> str:
        raise NotImplementedError("DataFusion backend does not support current_catalog")

    @property
    def current_database(self) -> str:
        raise NotImplementedError(
            "DataFusion backend does not support current_database"
        )

    def list_catalogs(self, like: str | None = None) -> list[str]:
        code = (
            sg.select(C.table_catalog)
            .from_(sg.table("tables", db="information_schema"))
            .distinct()
        ).sql()
        result = self.con.sql(code).to_pydict()
        return self._filter_with_like(result["table_catalog"], like)

    def create_catalog(self, name: str, force: bool = False) -> None:
        with self._safe_raw_sql(
            sge.Create(kind="DATABASE", this=sg.to_identifier(name), exists=force)
        ):
            pass

    def drop_catalog(self, name: str, force: bool = False) -> None:
        raise com.UnsupportedOperationError(
            "DataFusion does not support dropping databases"
        )

    def list_databases(
        self, like: str | None = None, catalog: str | None = None
    ) -> list[str]:
        return self._filter_with_like(
            self.con.catalog(catalog if catalog is not None else "datafusion").names(),
            like=like,
        )

    def create_database(
        self, name: str, catalog: str | None = None, force: bool = False
    ) -> None:
        # not actually a table, but this is how sqlglot represents schema names
        db_name = sg.table(name, db=catalog)
        with self._safe_raw_sql(sge.Create(kind="SCHEMA", this=db_name, exists=force)):
            pass

    def drop_database(
        self, name: str, catalog: str | None = None, force: bool = False
    ) -> None:
        db_name = sg.table(name, db=catalog)
        with self._safe_raw_sql(sge.Drop(kind="SCHEMA", this=db_name, exists=force)):
            pass

    def list_tables(
        self,
        like: str | None = None,
        database: str | None = None,
    ) -> list[str]:
        """Return the list of table names in the current database.

        Parameters
        ----------
        like
            A pattern in Python's regex format.
        database
            Unused in the datafusion backend.

        Returns
        -------
        list[str]
            The list of the table names that match the pattern `like`.
        """
        return self._filter_with_like(self.con.tables(), like)

    def get_schema(
        self,
        table_name: str,
        *,
        catalog: str | None = None,
        database: str | None = None,
    ) -> sch.Schema:
        if catalog is not None:
            catalog = self.con.catalog(catalog)
        else:
            catalog = self.con.catalog()

        if database is not None:
            database = catalog.database(database)
        else:
            database = catalog.database()

        table = database.table(table_name)
        return sch.schema(table.schema)

    def register(
        self,
        source: str | Path | pa.Table | pa.RecordBatch | pa.Dataset | pd.DataFrame,
        table_name: str | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        """Register a data set with `table_name` located at `source`.

        Parameters
        ----------
        source
            The data source(s). Maybe a path to a file or directory of
            parquet/csv files, a pandas dataframe, or a pyarrow table, dataset
            or record batch.
        table_name
            The name of the table
        kwargs
            Datafusion-specific keyword arguments

        """
        import pandas as pd
        import pyarrow.dataset as ds

        table_name = table_name or gen_name("register")
        table_ident = str(sg.to_identifier(table_name, quoted=self.compiler.quoted))

        if isinstance(source, (str, Path)):
            first = str(source)
        elif isinstance(source, pa.Table):
            self.con.deregister_table(table_ident)
            self.con.register_record_batches(table_ident, [source.to_batches()])
            return self.table(table_name)
        elif isinstance(source, pa.RecordBatch):
            self.con.deregister_table(table_ident)
            self.con.register_record_batches(table_ident, [[source]])
            return self.table(table_name)
        elif isinstance(source, ds.Dataset):
            self.con.deregister_table(table_ident)
            self.con.register_dataset(table_ident, source)
            return self.table(table_name)
        elif isinstance(source, pd.DataFrame):
            output = pa.Table.from_pandas(source)
            output = output.drop(
                filter(
                    lambda col: col.startswith("__index_level_"), output.column_names
                )
            )
            return self.register(output, table_name, **kwargs)
        elif isinstance(source, ibis.expr.types.Table) and hasattr(
            source, "to_pyarrow_batches"
        ):
            self.con.deregister_table(table_ident)
            self.con.register_table_provider(table_ident, IbisTableProvider(source))
            return self.table(table_name)
        elif isinstance(source, ibis.expr.types.Expr) and hasattr(
            source, "to_pyarrow_batches"
        ):
            self.con.deregister_table(table_ident)
            self.con.register_record_batch_reader(
                table_ident, source.to_pyarrow_batches()
            )
            return self.table(table_name)
        elif isinstance(source, pa.RecordBatchReader):
            if "ordering" in kwargs:
                kwargs["sort_order"] = translate_sort(kwargs.pop("ordering"))

            self.con.deregister_table(table_ident)
            self.con.register_record_batch_reader(table_ident, source, **kwargs)
            return self.table(table_name)
        elif isinstance(source, Table):
            self.con.deregister_table(table_ident)
            self.con.register_table(table_ident, source)
            return self.table(table_name)
        elif isinstance(source, DataFrame):
            self.con.deregister_table(table_ident)
            self.con.register_dataframe(table_ident, source)
            return self.table(table_name)
        else:
            raise ValueError(f"Unknown `source` type {type(source)}")

        if first.startswith(("parquet://", "parq://")) or first.endswith(
            ("parq", "parquet")
        ):
            return self.read_parquet(source, table_name=table_name, **kwargs)
        elif first.startswith(("csv://", "txt://")) or first.endswith(
            ("csv", "tsv", "txt")
        ):
            return self.read_csv(source, table_name=table_name, **kwargs)
        else:
            self._register_failure()
            return None

    def register_table_provider(
        self,
        source: str | Path | pa.Table | pa.RecordBatch | pa.Dataset | pd.DataFrame,
        table_name: str | None = None,
    ):
        table_ident = str(sg.to_identifier(table_name, quoted=self.compiler.quoted))
        self.con.deregister_table(table_ident)
        self.con.register_table_provider(table_ident, IbisTableProvider(source))
        return self.table(table_name)

    def _register_failure(self):
        import inspect

        msg = ", ".join(
            m[0] for m in inspect.getmembers(self) if m[0].startswith("read_")
        )
        raise ValueError(
            f"Cannot infer appropriate read function for input, "
            f"please call one of {msg} directly"
        )

    def _register_in_memory_table(self, op: ops.InMemoryTable) -> None:
        name = op.name
        schema = op.schema

        self.con.deregister_table(name)
        if batches := op.data.to_pyarrow(schema).to_batches():
            self.con.register_record_batches(name, [batches])
        else:
            import pyarrow.dataset as ds

            empty_dataset = ds.dataset([], schema=schema.to_pyarrow())
            self.con.register_dataset(name=name, dataset=empty_dataset)

    def _register_in_memory_tables(self, expr: ir.Expr) -> None:
        if self.supports_in_memory_tables:
            for memtable in expr.op().find(ops.InMemoryTable):
                self._register_in_memory_table(memtable)

    def read_csv(
        self, path: str | Path, table_name: str | None = None, **kwargs: Any
    ) -> ir.Table:
        """Register a CSV file as a table in the current database.

        Parameters
        ----------
        path
            The data source. A string or Path to the CSV file.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to Datafusion loading function.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        path = normalize_filenames(path)
        table_name = table_name or gen_name("read_csv")
        kwargs.setdefault("file_extension", Path(path[0]).suffix)

        storage_options, is_connection_set = make_s3_connection()
        if is_connection_set:
            kwargs["storage_options"] = storage_options | kwargs.get(
                "storage_options", {}
            )

        if schema := kwargs.get("schema"):
            if isinstance(schema, ibis.Schema):
                kwargs["schema"] = schema.to_pyarrow()

        # Our other backends support overwriting views / tables when re-registering
        self.con.deregister_table(table_name)
        self.con.register_csv(table_name, path, **kwargs)
        return self.table(table_name)

    def read_parquet(
        self, path: str | Path, table_name: str | None = None, **kwargs: Any
    ) -> ir.Table:
        """Register a parquet file as a table in the current database.

        Parameters
        ----------
        path
            The data source.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to Datafusion loading function.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        path = normalize_filenames(path)
        table_name = table_name or gen_name("read_parquet")
        kwargs.setdefault("file_extension", Path(path[0]).suffix)

        # Our other backends support overwriting views / tables when reregistering
        self.con.deregister_table(table_name)
        self.con.register_parquet(table_name, path, **kwargs)
        return self.table(table_name)

    def read_delta(
        self, source_table: str | Path, table_name: str | None = None, **kwargs: Any
    ) -> ir.Table:
        """Register a Delta Lake table as a table in the current database.

        Parameters
        ----------
        source_table
            The data source. Must be a directory
            containing a Delta Lake table.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to deltalake.DeltaTable.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        source_table = normalize_filename(source_table)

        table_name = table_name or gen_name("read_delta")

        # Our other backends support overwriting views / tables when reregistering
        self.con.deregister_table(table_name)

        try:
            from deltalake import DeltaTable
        except ImportError as err:
            raise ImportError(
                "The deltalake extra is required to use the "
                "read_delta method. You can install it using pip:\n\n"
                "pip install 'ibis-framework[deltalake]'\n"
            ) from err

        delta_table = DeltaTable(source_table, **kwargs)

        return self.register(delta_table.to_pyarrow_dataset(), table_name=table_name)

    def read_record_batches(self, source, table_name=None):
        table_name = table_name or gen_name("read_record_batches")
        table_ident = str(sg.to_identifier(table_name, quoted=self.compiler.quoted))
        self.con.deregister_table(table_ident)
        self.con.register_record_batch_reader(table_ident, source.cast(source.schema))
        return self.table(table_name)

    def to_pyarrow_batches(
        self,
        expr: ir.Expr,
        *,
        chunk_size: int = 1_000_000,
        **kwargs: Any,
    ) -> pa.ipc.RecordBatchReader:
        return self._to_pyarrow_batches(expr, chunk_size=chunk_size, **kwargs)

    def _to_pyarrow_batches(
        self,
        expr: ir.Expr,
        *,
        chunk_size: int = 1_000_000,
        **kwargs: Any,
    ):
        pa = self._import_pyarrow()
        self._register_udfs(expr)
        self._register_in_memory_tables(expr)
        table_expr = expr.as_table()
        raw_sql = self.compile(table_expr, **kwargs)
        frame = self.con.sql(raw_sql)
        schema = table_expr.schema()
        pyarrow_schema = schema.to_pyarrow()
        struct_schema = schema.as_struct().to_pyarrow()

        def make_gen():
            yield from (
                # convert the renamed + cast columns into a record batch
                pa.RecordBatch.from_struct_array(
                    # rename columns to match schema because datafusion lowercases things
                    pa.RecordBatch.from_arrays(
                        batch.to_pyarrow().columns, schema=pyarrow_schema
                    )
                    # cast the struct array to the desired types to work around
                    # https://github.com/apache/arrow-datafusion-python/issues/534
                    .to_struct_array()
                    .cast(struct_schema)
                )
                for batch in frame.execute_stream()
            )

        return pa.ipc.RecordBatchReader.from_batches(
            pyarrow_schema,
            make_gen(),
        )

    def to_pyarrow(self, expr: ir.Expr, **kwargs: Any) -> pa.Table:
        batch_reader = self._to_pyarrow_batches(expr, **kwargs)
        arrow_table = batch_reader.read_all()
        return expr.__pyarrow_result__(arrow_table)

    def execute(self, expr: ir.Expr, **kwargs: Any):
        batch_reader = self._to_pyarrow_batches(expr, **kwargs)
        return expr.__pandas_result__(
            batch_reader.read_pandas(timestamp_as_object=True)
        )

    def create_table(
        self,
        name: str,
        obj: pd.DataFrame | pa.Table | ir.Table | None = None,
        *,
        schema: sch.Schema | None = None,
        database: str | None = None,
        temp: bool = False,
        overwrite: bool = False,
    ):
        """Create a table in Datafusion.

        Parameters
        ----------
        name
            Name of the table to create
        obj
            The data with which to populate the table; optional, but at least
            one of `obj` or `schema` must be specified
        schema
            The schema of the table to create; optional, but at least one of
            `obj` or `schema` must be specified
        database
            The name of the database in which to create the table; if not
            passed, the current database is used.
        temp
            Create a temporary table
        overwrite
            If `True`, replace the table if it already exists, otherwise fail
            if the table exists

        """
        if obj is None and schema is None:
            raise ValueError("Either `obj` or `schema` must be specified")

        properties = []

        if temp:
            properties.append(sge.TemporaryProperty())

        quoted = self.compiler.quoted

        if obj is not None:
            if not isinstance(obj, ir.Expr):
                table = ibis.memtable(obj, schema=schema)
            else:
                table = obj

            self._run_pre_execute_hooks(table)
            compiler = self.compiler

            relname = "_"
            query = sg.select(
                *(
                    compiler.cast(
                        sg.column(col, table=relname, quoted=quoted), dtype
                    ).as_(col, quoted=quoted)
                    if not isinstance(dtype, dt.LargeString)
                    else compiler.f.arrow_cast(
                        sg.column(col, table=relname, quoted=quoted), "LargeUtf8"
                    ).as_(col, quoted=quoted)
                    for col, dtype in table.schema().items()
                )
            ).from_(
                compiler.to_sqlglot(table).subquery(
                    sg.to_identifier(relname, quoted=quoted)
                )
            )
        else:
            query = None

        table_ident = sg.to_identifier(name, quoted=quoted)
        if query is None:
            column_defs = [
                sge.ColumnDef(
                    this=sg.to_identifier(colname, quoted=quoted),
                    kind=self.compiler.type_mapper.from_ibis(typ),
                    constraints=(
                        None
                        if typ.nullable
                        else [sge.ColumnConstraint(kind=sge.NotNullColumnConstraint())]
                    ),
                )
                for colname, typ in (schema or table.schema()).items()
            ]
            target = sge.Schema(this=table_ident, expressions=column_defs)
        else:
            target = table_ident

        create_stmt = sge.Create(
            kind="TABLE",
            this=target,
            properties=sge.Properties(expressions=properties),
            expression=query,
            replace=overwrite,
        )

        with self._safe_raw_sql(create_stmt):
            pass

        return self.table(name, database=database)

    def truncate_table(
        self, name: str, database: str | None = None, schema: str | None = None
    ) -> None:
        """Delete all rows from a table.

        Parameters
        ----------
        name
            Table name
        database
            Database name
        schema
            Schema name

        """
        # datafusion doesn't support `TRUNCATE TABLE` so we use `DELETE FROM`
        #
        # however datafusion as of 34.0.0 doesn't implement DELETE DML yet
        table_loc = self._warn_and_create_table_loc(database, schema)
        catalog, db = self._to_catalog_db_tuple(table_loc)

        ident = sg.table(name, db=db, catalog=catalog).sql(self.name)
        with self._safe_raw_sql(sge.delete(ident)):
            pass

    def to_parquet(
        self,
        expr: ir.Table,
        path: str | Path,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        **kwargs: Any,
    ) -> None:
        self._import_pyarrow()
        import pyarrow.parquet as pq

        with expr.to_pyarrow_batches(params=params) as batch_reader:
            with pq.ParquetWriter(path, batch_reader.schema, **kwargs) as writer:
                for batch in batch_reader:
                    writer.write_batch(batch)


@contextlib.contextmanager
def _create_and_drop_memtable(_conn, table_name, tmp_name, overwrite):
    """Workaround inability to overwrite tables in dataframe API.

    DataFusion has helper methods for loading in-memory data, but these methods
    don't allow overwriting tables.
    The SQL interface allows creating tables from existing tables, so we register
    the data as a table using the dataframe API, then run a

    CREATE [OR REPLACE] TABLE table_name AS SELECT * FROM in_memory_thing

    and that allows us to toggle the overwrite flag.
    """
    src = sge.Create(
        this=table_name,
        kind="TABLE",
        expression=sg.select("*").from_(tmp_name),
        replace=overwrite,
    )

    yield

    _conn.raw_sql(src)
    _conn.drop_table(tmp_name)


@lazy_singledispatch
def _read_in_memory(
    source: Any, table_name: str, _conn: Backend, overwrite: bool = False
):
    raise NotImplementedError("No support for source or imports missing")


@_read_in_memory.register(dict)
def _pydict(source, table_name, _conn, overwrite: bool = False):
    tmp_name = gen_name("pydict")
    with _create_and_drop_memtable(_conn, table_name, tmp_name, overwrite):
        _conn.con.from_pydict(source, name=tmp_name)


@_read_in_memory.register("polars.DataFrame")
def _polars(source, table_name, _conn, overwrite: bool = False):
    tmp_name = gen_name("polars")
    with _create_and_drop_memtable(_conn, table_name, tmp_name, overwrite):
        _conn.con.from_polars(source, name=tmp_name)


@_read_in_memory.register("polars.LazyFrame")
def _polars(source, table_name, _conn, overwrite: bool = False):
    tmp_name = gen_name("polars")
    with _create_and_drop_memtable(_conn, table_name, tmp_name, overwrite):
        _conn.con.from_polars(source.collect(), name=tmp_name)


@_read_in_memory.register("pyarrow.Table")
def _pyarrow_table(source, table_name, _conn, overwrite: bool = False):
    tmp_name = gen_name("pyarrow")
    with _create_and_drop_memtable(_conn, table_name, tmp_name, overwrite):
        _conn.con.from_arrow(source, name=tmp_name)


@_read_in_memory.register("pyarrow.RecordBatchReader")
def _pyarrow_rbr(source, table_name, _conn, overwrite: bool = False):
    tmp_name = gen_name("pyarrow")
    with _create_and_drop_memtable(_conn, table_name, tmp_name, overwrite):
        _conn.con.from_arrow(source.read_all(), name=tmp_name)


@_read_in_memory.register("pyarrow.RecordBatch")
def _pyarrow_rb(source, table_name, _conn, overwrite: bool = False):
    tmp_name = gen_name("pyarrow")
    with _create_and_drop_memtable(_conn, table_name, tmp_name, overwrite):
        _conn.con.register_record_batches(tmp_name, [[source]])


@_read_in_memory.register("pyarrow.dataset.Dataset")
def _pyarrow_rb(source, table_name, _conn, overwrite: bool = False):
    tmp_name = gen_name("pyarrow")
    with _create_and_drop_memtable(_conn, table_name, tmp_name, overwrite):
        _conn.con.register_dataset(tmp_name, source)


@_read_in_memory.register("pandas.DataFrame")
def _pandas(source: pd.DataFrame, table_name, _conn, overwrite: bool = False):
    tmp_name = gen_name("pandas")
    with _create_and_drop_memtable(_conn, table_name, tmp_name, overwrite):
        _conn.con.from_pandas(source, name=tmp_name)



import contextlib
import functools
import gzip
import hashlib
import shutil
import tarfile
from pathlib import Path
from tempfile import TemporaryDirectory

import toolz
from attr import (
    field,
    frozen,
)
from attr.validators import (
    instance_of,
)


# paths that uv build injects that don't impact package functionality
uv_sdist_omit_suffixes = ("PKG-INFO", ".gitignore")


@toolz.curried.excepts(ValueError)
def try_path_relative_to(from_, to_):
    return Path(from_).relative_to(to_)


def uv_sdist_member_filter(member):
    return not any(member.name.endswith(suffix) for suffix in uv_sdist_omit_suffixes)


def gunzip_path(from_, to_):
    with gzip.GzipFile(from_, "rb") as gfh:
        with to_.open("wb") as fh:
            shutil.copyfileobj(gfh, fh)
    return to_


def gzip_path(from_, to_):
    with from_.open("rb") as fh:
        with gzip.GzipFile(to_, "wb") as gfh:
            shutil.copyfileobj(fh, gfh)
    return to_


def copy_path(from_, to_):
    with from_.open("rb") as from_fh:
        with to_.open("wb") as to_fh:
            shutil.copyfileobj(from_fh, to_fh)


def tar_append(tar_path, append_path, **kwargs):
    with tarfile.TarFile(tar_path, mode="a") as tf:
        tf.add(append_path, **kwargs)


def get_root_dir(tgz_path):
    with tarfile.TarFile.gzopen(tgz_path) as tf:
        (name, *rest) = tf.getnames()
        (root_dir, *_) = Path(name).parts
        assert root_dir
        assert all(other.startswith(root_dir) for other in rest)
        return Path(root_dir)


def calc_tgz_content_hexdigest(path, member_filter=uv_sdist_member_filter):
    # ignore metadata like permissions and modification time
    from xorq.common.utils.dask_normalize.dask_normalize_utils import (
        file_digest,
    )

    with tarfile.TarFile.gzopen(path) as tf:
        dct = {
            name: file_digest(fh, hashlib.md5)
            for name, fh in (
                (member.name, tf.extractfile(member))
                for member in tf.getmembers()
                if not member.isdir() and member_filter(member)
            )
            if fh
        }
        print(tuple(sorted(dct.items())))
        md5 = hashlib.md5()
        for _key, value in sorted(dct.items()):
            # can't use key: top level dir changes shouldn't impact hash
            # md5.update(key.encode("ascii"))
            md5.update(value.encode("ascii"))
        return md5.hexdigest()


@frozen
class TGZProxy:
    tgz_path = field(validator=instance_of(Path), converter=Path)

    valid_full_suffixes = (".tar.gz", ".tgz")

    def __attrs_post_init__(self):
        full_suffix = "".join(self.tgz_path.suffixes)
        assert any(full_suffix.endswith(suffix) for suffix in self.valid_full_suffixes)

    @property
    @functools.cache
    def root_dir(self):
        return get_root_dir(self.tgz_path)

    def toplevel_name_exists(self, name):
        with self.open() as tf:
            return any(
                other and str(other) == name
                for other in (
                    try_path_relative_to(other, self.root_dir)
                    for other in tf.getnames()
                )
            )

    @contextlib.contextmanager
    def open(self):
        with tarfile.TarFile.gzopen(self.tgz_path) as tf:
            yield tf

    @contextlib.contextmanager
    def open_member(self, member_path):
        # this is a bytes stream
        with self.open() as tf:
            yield tf.extractfile(str(member_path))

    @contextlib.contextmanager
    def open_toplevel_member(self, member_path):
        # this is a bytes stream
        with self.open_member(self.root_dir.joinpath(member_path)) as fh:
            yield fh

    def extract_toplevel_name(self, name, dest):
        dest = Path(dest)
        with dest.open("wb") as ofh:
            with self.open_toplevel_member(name) as ifh:
                ofh.write(ifh.read())
        return dest

    @property
    def members(self):
        with self.open() as tf:
            return tf.getmembers()

    def extract_toplevel(self, dest):
        dest = Path(dest).absolute()
        names = (
            info["name"]
            for info in (member.get_info() for member in self.members)
            if info["type"] != b"5"
        )
        for name in names:
            relpath = Path(name).relative_to(self.root_dir)
            destpath = dest.joinpath(relpath)
            destpath.parent.mkdir(exist_ok=True, parents=True)
            self.extract_toplevel_name(str(relpath), destpath)
        return dest


@frozen
class TGZAppender:
    tgz_path = field(validator=instance_of(Path), converter=Path)
    append_path = field(validator=instance_of(Path), converter=Path)
    kwargs_tuple = field(validator=instance_of(tuple))

    @property
    def kwargs(self):
        return dict(self.kwargs_tuple)

    @property
    @functools.cache
    def root_dir(self):
        return get_root_dir(self.tgz_path)

    @property
    @functools.cache
    def _tmpdir(self):
        return TemporaryDirectory()

    @property
    def tmpdir(self):
        return Path(self._tmpdir.name)

    @property
    @functools.cache
    def gunzipped_path(self):
        gunzipped_path = self.tmpdir.joinpath("gunzipped.tar")
        gunzip_path(self.tgz_path, gunzipped_path)
        return gunzipped_path

    @property
    @functools.cache
    def appended_path(self):
        appended_path = self.tmpdir.joinpath("appended.tar")
        copy_path(self.gunzipped_path, appended_path)
        tar_append(appended_path, self.append_path, **self.kwargs)
        return appended_path

    @property
    @functools.cache
    def appended_tgz_path(self):
        appended_tgz_path = self.tmpdir.joinpath(self.tgz_path.name)
        gzip_path(self.appended_path, appended_tgz_path)
        return appended_tgz_path

    @classmethod
    def append_toplevel(cls, tgz_path, append_path, suffix=".bak", **kwargs):
        # write to same location as the original tgz_path
        append_path = Path(append_path)
        arcname = str(TGZProxy(tgz_path).root_dir.joinpath(append_path.name))
        kwargs_tuple = tuple(({"arcname": arcname} | kwargs).items())
        self = cls(tgz_path, append_path, kwargs_tuple=kwargs_tuple)
        # make sure the append is successful
        self.appended_tgz_path
        renamed = self.tgz_path.with_name(self.tgz_path.name + suffix)
        self.tgz_path.rename(renamed)
        self.appended_tgz_path.rename(self.tgz_path)
        return tgz_path, renamed



"""Tests for Arrow IPC streaming in CLI."""

import io
import re

import pyarrow as pa
import pytest

import xorq.api as xo
from xorq.common.utils.process_utils import subprocess_run


def test_to_pyarrow_stream():
    """Test writing Arrow IPC stream to a buffer."""
    # Create test data
    table = pa.table({"a": [1, 2, 3], "b": [4, 5, 6]})
    expr = xo.connect().read_record_batches(table)
    # Write to buffer
    buf = io.BytesIO()
    xo.to_pyarrow_stream(expr, sink=buf)

    # Read back
    buf.seek(0)
    result = pa.ipc.open_stream(buf).read_all()
    assert result.equals(table)


def test_read_pyarrow_stream():
    """Test reading Arrow IPC stream from a buffer."""
    # Create test data
    table = pa.table({"x": [10, 20, 30], "y": [40, 50, 60]})

    # Write to buffer
    buf = io.BytesIO()
    writer = pa.ipc.new_stream(buf, table.schema)
    for batch in table.to_batches():
        writer.write_batch(batch)
    writer.close()

    # Read back using our function
    buf.seek(0)
    result = xo.read_pyarrow_stream(buf).to_pyarrow()
    assert result.equals(table)


@pytest.mark.slow(level=1)
def test_run_command_arrow_output(tmp_path, fixture_dir):
    """Test running a build with Arrow output format."""
    target_dir = tmp_path / "build"
    script_path = fixture_dir / "pipeline.py"

    # Build the expression
    build_args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(target_dir),
    ]
    (returncode, stdout, stderr) = subprocess_run(build_args)
    assert returncode == 0, stderr

    if match := re.search(f"{target_dir}/([0-9a-f]+)", stdout.decode("ascii")):
        output_path = tmp_path / "test.arrow"
        expression_path = match.group()

        # Run with arrow output
        test_args = [
            "xorq",
            "run",
            expression_path,
            "--output-path",
            str(output_path),
            "--format",
            "arrow",
        ]
        (returncode, _, stderr) = subprocess_run(test_args)
        assert returncode == 0, stderr
        assert output_path.exists()
        assert output_path.stat().st_size > 0

        # Verify we can read the arrow file
        with open(output_path, "rb") as f:
            reader = pa.ipc.open_stream(f)
            table = reader.read_all()
            assert len(table) > 0

    else:
        raise AssertionError("No expression hash")


@pytest.mark.slow(level=1)
def test_run_command_arrow_output_stdout(tmp_path, fixture_dir):
    """Test running a build with Arrow output to stdout."""
    target_dir = tmp_path / "build"
    script_path = fixture_dir / "pipeline.py"

    # Build the expression
    build_args = [
        "xorq",
        "build",
        str(script_path),
        "--expr-name",
        "expr",
        "--builds-dir",
        str(target_dir),
    ]
    (returncode, stdout, stderr) = subprocess_run(build_args)
    assert returncode == 0, stderr

    if match := re.search(f"{target_dir}/([0-9a-f]+)", stdout.decode("ascii")):
        expression_path = match.group()

        # Run with arrow output to stdout
        test_args = [
            "xorq",
            "run",
            expression_path,
            "--output-path",
            "-",
            "--format",
            "arrow",
        ]
        (returncode, stdout, stderr) = subprocess_run(test_args)
        assert returncode == 0, stderr
        assert len(stdout) > 0

        # Verify the stdout contains valid Arrow IPC data
        buf = io.BytesIO(stdout)
        reader = pa.ipc.open_stream(buf)
        table = reader.read_all()
        assert len(table) > 0

    else:
        raise AssertionError("No expression hash")



"""Deferred scikit-learn metrics evaluation for xorq."""

import functools
import inspect
from types import MappingProxyType
from typing import Callable

import numpy as np
import toolz
from attr import (
    field,
    frozen,
)
from attr.validators import (
    instance_of,
    optional,
)

import xorq.expr.datatypes as dt
import xorq.expr.udf as udf
from xorq.common.utils.attr_utils import (
    convert_sorted_kwargs_tuple,
    validate_kwargs_tuple,
)
from xorq.common.utils.name_utils import make_name
from xorq.expr.ml.enums import ResponseMethod


@functools.cache
def _build_known_scorer_funcs():
    """Set of known scorer functions.

    Used to check whether a bare callable is a known sklearn scorer function
    (e.g. accuracy_score, mean_squared_error) vs an unknown callable
    (e.g. confusion_matrix, custom fn).
    """
    from sklearn.metrics import get_scorer, get_scorer_names

    return frozenset(get_scorer(name)._score_func for name in get_scorer_names())


def _default_scorer_for_model(model):
    """Get the default scorer based on model type.

    Parameters
    ----------
    model : estimator
        An sklearn estimator instance (fitted or unfitted).  Only the
        class hierarchy (``ClassifierMixin``, ``RegressorMixin``,
        ``ClusterMixin``) is inspected, so a fitted model is never
        required.

    Returns
    -------
    scorer : _BaseScorer
        An sklearn scorer object appropriate for the model type.

    Raises
    ------
    ValueError
        If model type is not recognized.
    """
    from sklearn.base import ClassifierMixin, ClusterMixin, RegressorMixin
    from sklearn.metrics import (
        accuracy_score,
        adjusted_rand_score,
        make_scorer,
        r2_score,
    )

    match model:
        case ClusterMixin():
            return make_scorer(adjusted_rand_score)
        case ClassifierMixin():
            return make_scorer(accuracy_score)
        case RegressorMixin():
            return make_scorer(r2_score)
        case _:
            raise ValueError(
                f"Cannot determine default scorer for model type {type(model).__name__}. "
                "Please specify a scorer explicitly."
            )


@frozen
class Scorer:
    """Normalized representation of a scorer.

    Encapsulates all detection logic: given any input (str, _BaseScorer,
    bare callable, or None), resolves to a consistent set of properties.
    """

    metric_fn: Callable = field(validator=instance_of(Callable))
    sign: int = field(validator=instance_of(int))
    kwargs: tuple = field(
        factory=tuple,
        converter=convert_sorted_kwargs_tuple,
    )
    response_method: str = field(
        validator=instance_of(str), default=ResponseMethod.PREDICT
    )

    @classmethod
    def from_model(cls, model):
        """Create a Scorer from a model's default metric.

        Parameters
        ----------
        model : estimator
            An sklearn estimator instance (fitted or unfitted).  Only the
            class hierarchy is inspected to determine the default scorer.

        Returns
        -------
        Scorer
            A Scorer using the model's default metric (accuracy for
            classifiers, r2 for regressors, adjusted_rand for clusterers).
        """
        scorer_obj = _default_scorer_for_model(model)
        return cls._from_scorer_obj(scorer_obj)

    @classmethod
    def from_spec(cls, scorer):
        """Normalize str | callable | _BaseScorer -> Scorer.

        Parameters
        ----------
        scorer : str, callable, _BaseScorer, or Scorer
            The scorer specification to normalize.

        Returns
        -------
        Scorer
            A normalized Scorer instance.
        """
        from sklearn.metrics import get_scorer
        from sklearn.metrics._scorer import _BaseScorer

        match scorer:
            case str():
                scorer_obj = get_scorer(scorer)
                return cls._from_scorer_obj(scorer_obj)

            case _BaseScorer():
                return cls._from_scorer_obj(scorer)

            case Scorer():
                return scorer  # already resolved

            case object(__call__=_):
                known = _build_known_scorer_funcs()
                if scorer in known:
                    # Known scorer function — bare callable = raw metric value
                    # sign=1, kwargs={} (function's own defaults apply;
                    # caller provides overrides via metric_kwargs)
                    return cls(
                        metric_fn=scorer,
                        sign=1,
                        kwargs={},
                        response_method=ResponseMethod.PREDICT,
                    )
                else:
                    raise ValueError(
                        f"Unknown callable {scorer.__name__!r} is not a known sklearn scorer function. "
                        f"Use make_scorer() to wrap custom metrics, or pass a scorer name string."
                    )

            case _:
                raise ValueError(
                    f"scorer must be a string, callable, _BaseScorer, or Scorer, "
                    f"got {type(scorer)}"
                )

    @classmethod
    def _from_scorer_obj(cls, scorer_obj):
        """Create a Scorer from an sklearn _BaseScorer object."""
        return cls(
            metric_fn=scorer_obj._score_func,
            sign=scorer_obj._sign,
            kwargs=scorer_obj._kwargs,
            response_method=cls._resolve_response_method(scorer_obj),
        )

    @staticmethod
    def _resolve_response_method(scorer_obj):
        """Extract the response method string from a scorer object."""
        raw = scorer_obj._response_method
        match raw:
            case (method, *_):
                return method
            case str(method):
                return method
            case _:
                raise ValueError(f"Unexpected _response_method: {raw}")


@functools.cache
def _build_known_non_scorer_metric_fns():
    """Set of known non-scorer metric functions.

    Computed as all public functions in ``sklearn.metrics`` minus the
    scorer functions from ``_build_known_scorer_funcs()``, excluding
    pairwise/scorer utilities by module and unsupported metrics by name.
    """
    import sklearn.metrics

    _EXCLUDED_NAMES = frozenset(
        {
            "auc",  # helper for curve coords, not (y_true, y_pred)
            "classification_report",  # returns string
            "consensus_score",  # biclustering, different signature
            "multilabel_confusion_matrix",  # returns 3D array (n_classes, 2, 2)
            "precision_recall_fscore_support",  # returns variable-length tuple
        }
    )

    all_metric_fns = frozenset(
        obj
        for name, obj in sklearn.metrics.__dict__.items()
        if not name.startswith("_")
        and name not in _EXCLUDED_NAMES
        and inspect.isfunction(obj)
        and not obj.__module__.startswith(
            ("sklearn.metrics.pairwise", "sklearn.metrics._scorer")
        )
    )
    return all_metric_fns - _build_known_scorer_funcs()


@functools.cache
def _build_metric_return_types():
    """Registry of non-scalar return types for known sklearn metrics.

    Maps metric functions to their ``dt.DataType`` so that
    ``deferred_sklearn_metric`` can auto-resolve ``return_type``
    when the caller doesn't provide one explicitly.
    """
    from sklearn.metrics import (
        class_likelihood_ratios,
        confusion_matrix,
        det_curve,
        homogeneity_completeness_v_measure,
        pair_confusion_matrix,
        precision_recall_curve,
        roc_curve,
        silhouette_samples,
    )

    return MappingProxyType(
        {
            # Tuple of scalars -> Struct
            class_likelihood_ratios: dt.Struct(
                {
                    "positive_likelihood_ratio": dt.float64,
                    "negative_likelihood_ratio": dt.float64,
                }
            ),
            homogeneity_completeness_v_measure: dt.Struct(
                {
                    "homogeneity": dt.float64,
                    "completeness": dt.float64,
                    "v_measure": dt.float64,
                }
            ),
            # Confusion matrices -> Array(Array(int64))
            confusion_matrix: dt.Array(dt.Array(dt.int64)),
            pair_confusion_matrix: dt.Array(dt.Array(dt.int64)),
            # Curves -> Struct of arrays
            roc_curve: dt.Struct(
                {
                    "fpr": dt.Array(dt.float64),
                    "tpr": dt.Array(dt.float64),
                    "thresholds": dt.Array(dt.float64),
                }
            ),
            precision_recall_curve: dt.Struct(
                {
                    "precision": dt.Array(dt.float64),
                    "recall": dt.Array(dt.float64),
                    "thresholds": dt.Array(dt.float64),
                }
            ),
            det_curve: dt.Struct(
                {
                    "fpr": dt.Array(dt.float64),
                    "fnr": dt.Array(dt.float64),
                    "thresholds": dt.Array(dt.float64),
                }
            ),
            # Per-sample metrics -> Array(float64)
            silhouette_samples: dt.Array(dt.float64),
        }
    )


def _validate_str_or_tuple_of_str(instance, attribute, value):
    """Validate value is str or tuple of str."""
    match value:
        case str():
            pass
        case tuple() if all(isinstance(v, str) for v in value):
            pass
        case _:
            raise TypeError(
                f"{attribute.name} must be a str or tuple of str, got {type(value)}"
            )


@frozen
class MetricComputation:
    target = field(validator=_validate_str_or_tuple_of_str)
    pred = field(validator=_validate_str_or_tuple_of_str)
    metric_fn: Callable = field(validator=instance_of(Callable))
    sign: int = field(validator=optional(instance_of(int)), default=None)
    metric_kwargs_tuple: tuple = field(
        factory=tuple,
        validator=validate_kwargs_tuple,
        converter=convert_sorted_kwargs_tuple,
    )
    return_type = field(validator=instance_of(dt.DataType), default=dt.float64)
    name = field(validator=optional(instance_of(str)), default=None)

    def __attrs_post_init__(self):
        if self.name is None:
            object.__setattr__(
                self,
                "name",
                make_name(
                    prefix=f"metric_{self.metric_fn.__name__}",
                    to_tokenize=self.metric_kwargs_tuple,
                ),
            )

    @property
    def metric_kwargs(self):
        return dict(self.metric_kwargs_tuple)

    @property
    def __name__(self):
        """Return the name of the metric function for UDF registration."""
        return self.metric_fn.__name__

    @property
    def __module__(self):
        """Return the module of the metric function for UDF registration."""
        return self.metric_fn.__module__

    @staticmethod
    def _normalize_columns(value, name):
        """Normalize str or tuple[str] to tuple[str].

        Parameters
        ----------
        value : str or tuple of str
            Column name(s) to normalize.
        name : str
            Parameter name for error messages.

        Returns
        -------
        tuple of str
            Normalized column names as a tuple.
        """
        match value:
            case str():
                return (value,)
            case tuple():
                return value
            case _:
                raise TypeError(
                    f"{name} must be a str or tuple of str, got {type(value)}"
                )

    @property
    def _target_columns(self):
        return self._normalize_columns(self.target, "target")

    @property
    def _pred_columns(self):
        return self._normalize_columns(self.pred, "pred")

    @property
    def _prepare_target(self):
        """Extract the target (first metric argument) from a DataFrame.

        str  -> single column (Series)
        tuple -> multiple columns as 2D ndarray
        """
        match self.target:
            case str():
                return lambda df: df[self.target]
            case tuple():
                return lambda df: df[list(self.target)].values
            case _:
                raise TypeError(
                    f"target must be a str or tuple of str, got {type(self.target)}"
                )

    @property
    def _prepare_predictions(self):
        """Extract predictions (second metric argument) from a DataFrame.

        Dispatch is on (pred type, target type):

        (str, str)   -> single column with positive-class extraction
                        for binary (n, 2) probabilities
        (str, tuple) -> single column, raw (no extraction); handles both
                        scalar columns (clustering) and array columns
                        (multilabel)
        (tuple, any) -> multiple columns as 2D ndarray
        """
        match (self.pred, self.target):
            case (str(), str()):
                return lambda df: self._extract_positive_class_proba(
                    self._coerce_ndarray(df[self.pred])
                )
            case (str(), tuple()):
                return lambda df: self._coerce_ndarray(df[self.pred])
            case (tuple(), _):
                return lambda df: df[list(self.pred)].values
            case _:
                raise TypeError(
                    f"pred must be a str or tuple of str, got {type(self.pred)}"
                )

    @property
    def _apply_sign(self):
        """Dispatch for sign application to a metric result."""
        match self.sign:
            case None:
                return lambda result: result
            case int():
                return lambda result: self.sign * result
            case _:
                raise TypeError(f"sign must be None or int, got {type(self.sign)}")

    @property
    def _convert_result_for_udaf(self):
        """Convert a raw sklearn metric result for the UDAF infrastructure.

        dt.Struct  -> dict keyed by field names (tuple of scalars or arrays)
        dt.Array   -> list (ndarray.tolist() for matrices / per-sample)
        dt.Float64 -> passthrough
        """
        match self.return_type:
            case dt.Struct() as s:
                names = s.names
                return lambda raw: dict(zip(names, raw))
            case dt.Array():
                return lambda raw: raw.tolist()
            case dt.Float64():
                return lambda raw: raw
            case _:
                raise TypeError(f"Unsupported return_type {self.return_type}")

    def __call__(self, df):
        target = self._prepare_target(df)
        predictions = self._prepare_predictions(df)
        result = self.metric_fn(target, predictions, **self.metric_kwargs)
        return self._apply_sign(self._convert_result_for_udaf(result))

    def on_expr(self, expr):
        schema = expr.select((*self._target_columns, *self._pred_columns)).schema()
        metric_udaf = udf.agg.pandas_df(
            fn=self,
            schema=schema,
            return_type=self.return_type,
            name=self.name,
        )
        return metric_udaf.on_expr(expr)

    @staticmethod
    def _coerce_ndarray(predictions):
        """Coerce a column value into a numpy ndarray.

        Handles scalar Series, array-valued Series (np.vstack), and
        raw ndarrays.
        """
        import pandas as pd

        match predictions:
            case pd.Series() as series if len(series) > 0 and isinstance(
                series.iloc[0], (np.ndarray, list, tuple)
            ):
                return np.vstack(series.values)
            case pd.Series() as series:
                return series.values
            case np.ndarray():
                return predictions
            case _:
                return predictions

    @staticmethod
    def _extract_positive_class_proba(y_pred):
        """Extract positive class probabilities for binary classification."""
        match (y_pred.ndim, y_pred.shape):
            case (2, (_, 2)):
                return y_pred[:, 1]
            case (2, (_, 1)):
                return y_pred[:, 0]
            case _:
                return y_pred


@toolz.curry
def deferred_sklearn_metric(
    expr,
    target,
    pred,
    metric,
    *,
    metric_kwargs=(),
    return_type=dt.float64,
    name=None,
):
    """Compute sklearn metrics on expressions.

    This function expects that predictions have already been added to the
    expression (via fitted_pipeline.predict(), predict_proba(), or
    decision_function()).

    Parameters
    ----------
    expr : ibis.Expr
        Expression containing both target and prediction columns.
    target : str | tuple[str, ...]
        Name of the target column, or a tuple of column names for metrics
        that expect a feature matrix (e.g. clustering metrics like
        calinski_harabasz_score).
    pred : str | tuple[str, ...]
        Name of the prediction column (e.g., "predict", "predict_proba"),
        or a tuple of column names for multi-column predictions.
    metric : str | _BaseScorer | Callable | Scorer
        The metric to compute.  Accepted forms:

        - **str** -- scorer name from ``sklearn.metrics.get_scorer_names()``
          (e.g. ``"accuracy"``, ``"neg_mean_squared_error"``).
        - **_BaseScorer** -- an sklearn scorer object (e.g. from
          ``make_scorer()``).
        - **Scorer** -- an already-resolved ``Scorer`` instance.
        - **callable** -- a known sklearn metric function.  Must be either
          a known scorer function (e.g. ``accuracy_score``) or a known
          non-scorer metric (e.g. ``cohen_kappa_score``,
          ``confusion_matrix``).  Unknown callables are rejected.
    metric_kwargs : dict, optional
        Additional kwargs to pass to the metric function.
    return_type : dt.DataType, optional
        Return type for the metric.  Auto-detected for non-scalar metrics
        in the registry; defaults to ``dt.float64`` for scalar metrics.
    name : str, optional
        Custom name for the UDF.

    Returns
    -------
    deferred_metric : ibis.Expr
        Deferred expression that computes the metric when executed.

    Examples
    --------
    >>> from sklearn.metrics import accuracy_score, cohen_kappa_score, confusion_matrix
    >>>
    >>> # Scorer name string
    >>> acc = deferred_sklearn_metric(
    ...     expr_with_preds, target="target", pred="predict",
    ...     metric="accuracy",
    ... )
    >>>
    >>> # Known scorer callable
    >>> acc = deferred_sklearn_metric(
    ...     expr_with_preds, target="target", pred="predict",
    ...     metric=accuracy_score,
    ... )
    >>>
    >>> # Non-scorer metric
    >>> kappa = deferred_sklearn_metric(
    ...     expr_with_preds, target="target", pred="predict",
    ...     metric=cohen_kappa_score,
    ... )
    >>>
    >>> # Non-scalar metric (return_type auto-detected)
    >>> cm = deferred_sklearn_metric(
    ...     expr_with_preds, target="target", pred="predict",
    ...     metric=confusion_matrix,
    ... )
    """
    from sklearn.metrics._scorer import _BaseScorer

    # Helper to build MetricComputation from a Scorer
    def _from_scorer(scorer, metric_kwargs):
        merged_kwargs = dict(scorer.kwargs) | dict(metric_kwargs)
        return MetricComputation(
            target=target,
            pred=pred,
            metric_fn=scorer.metric_fn,
            sign=scorer.sign,
            metric_kwargs_tuple=merged_kwargs,
            return_type=return_type,
            name=name,
        ).on_expr(expr)

    # Helper to build MetricComputation from a non-scorer metric function
    def _from_non_scorer_metric_fn(metric_fn, metric_kwargs):
        resolved_return_type = _build_metric_return_types().get(metric_fn, return_type)
        return MetricComputation(
            target=target,
            pred=pred,
            metric_fn=metric_fn,
            sign=None,
            metric_kwargs_tuple=dict(metric_kwargs),
            return_type=resolved_return_type,
            name=name,
        ).on_expr(expr)

    match metric:
        case str() | _BaseScorer():
            return _from_scorer(Scorer.from_spec(metric), metric_kwargs)

        case Scorer():
            return _from_scorer(metric, metric_kwargs)

        case object(__call__=_):
            known_scorers = _build_known_scorer_funcs()
            known_non_scorer_metrics = _build_known_non_scorer_metric_fns()

            match (metric in known_scorers, metric in known_non_scorer_metrics):
                case (True, False):
                    return _from_scorer(Scorer.from_spec(metric), metric_kwargs)
                case (False, True):
                    return _from_non_scorer_metric_fn(metric, metric_kwargs)
                case _:
                    raise ValueError(
                        f"Unknown callable {metric.__name__!r}. "
                        f"Must be a known sklearn scorer function or a known "
                        f"non-scorer metric. Use a scorer name string or "
                        f"make_scorer() for custom metrics."
                    )

        case _:
            raise TypeError(
                f"metric must be a str, _BaseScorer, Scorer, or known callable, "
                f"got {type(metric)}"
            )


# sklearn curve functions return plain tuples — no field-name metadata — so
# the Struct field names in _build_metric_return_types are ours.  The Struct
# field order must match sklearn's positional return order because
# _convert_result_for_udaf zips names with the tuple: dict(zip(names, raw)).
#
# auc(x, y) requires x to be monotonic.  precision_recall_curve returns
# (precision, recall, thresholds) but recall is the monotonic axis, so auc
# needs (recall, precision) — the reverse of the Struct order.  We can't
# reorder the Struct without breaking the zip, so this map exists to resolve
# the correct (x, y) pair for each curve regardless of Struct field order.
_CURVE_FIELD_MAP = {
    frozenset({"fpr", "tpr", "thresholds"}): ("fpr", "tpr"),
    frozenset({"precision", "recall", "thresholds"}): ("recall", "precision"),
    frozenset({"fpr", "fnr", "thresholds"}): ("fpr", "fnr"),
}


def deferred_auc_from_curve(curve_expr):
    """Compute the area under a deferred curve metric.

    Parameters
    ----------
    curve_expr : ibis.Expr
        A deferred curve expression with a Struct return type,
        from ``deferred_sklearn_metric`` with ``roc_curve``,
        ``precision_recall_curve``, or ``det_curve``.

    Returns
    -------
    auc_expr : ibis.Expr
        A deferred float64 scalar expression.

    Examples
    --------
    >>> from sklearn.metrics import roc_curve
    >>> deferred_roc = deferred_sklearn_metric(
    ...     expr=preds, target="target", pred="scores",
    ...     metric=roc_curve,
    ... )
    >>> deferred_roc_auc = deferred_auc_from_curve(deferred_roc)
    """
    import pyarrow as pa
    from sklearn.metrics import auc

    # Validate type and extract field mapping
    curve_type = curve_expr.type()
    match curve_type:
        case dt.Struct():
            field_names = frozenset(curve_type.names)
            if (xy_fields := _CURVE_FIELD_MAP.get(field_names)) is None:
                raise ValueError(
                    f"Unrecognized curve fields {set(field_names)}. "
                    f"Expected fields from roc_curve, precision_recall_curve, "
                    f"or det_curve."
                )
            x_field, y_field = xy_fields
        case _:
            raise TypeError(
                f"Expected a Struct expression from a curve metric, got {curve_type}"
            )

    # Create UDF to compute AUC
    def _auc_fn(struct: curve_type) -> dt.float64:
        x = struct.field(x_field).values.to_pylist()
        y = struct.field(y_field).values.to_pylist()
        return pa.array([auc(x, y)], type=pa.float64())

    _auc_fn.__name__ = f"_auc_{x_field}_{y_field}"
    return udf.scalar.pyarrow(_auc_fn)(curve_expr)



from __future__ import annotations

import contextlib
import inspect
import typing
from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING, Any

import datafusion as df
import pyarrow as pa
import pyarrow.dataset as ds
import pyarrow_hotfix  # noqa: F401
import sqlglot as sg
import sqlglot.expressions as sge

import xorq.common.exceptions as com
import xorq.vendor.ibis.backends.sql.compilers as sc
import xorq.vendor.ibis.expr.datatypes as dt
import xorq.vendor.ibis.expr.operations as ops
import xorq.vendor.ibis.expr.schema as sch
import xorq.vendor.ibis.expr.types as ir
from xorq.vendor import ibis
from xorq.vendor.ibis import util
from xorq.vendor.ibis.backends import (
    CanCreateCatalog,
    CanCreateDatabase,
    CanCreateSchema,
    NoUrl,
)
from xorq.vendor.ibis.backends.sql import SQLBackend
from xorq.vendor.ibis.backends.sql.compilers.base import C
from xorq.vendor.ibis.common.dispatch import lazy_singledispatch
from xorq.vendor.ibis.expr.operations.udf import InputType
from xorq.vendor.ibis.formats.pyarrow import PyArrowSchema, PyArrowType
from xorq.vendor.ibis.util import deprecated, gen_name, normalize_filename


try:
    from datafusion import ExecutionContext as SessionContext
except ImportError:
    from datafusion import SessionContext

try:
    from datafusion import SessionConfig
except ImportError:
    SessionConfig = None

try:
    from datafusion import RuntimeConfig
except ImportError:
    RuntimeConfig = None

if TYPE_CHECKING:
    import pandas as pd
    import polars as pl


def as_nullable(dtype: dt.DataType) -> dt.DataType:
    """Recursively convert a possibly non-nullable datatype to a nullable one."""
    if dtype.is_struct():
        return dtype.copy(
            fields={name: as_nullable(typ) for name, typ in dtype.items()},
            nullable=True,
        )
    elif dtype.is_array():
        return dtype.copy(value_type=as_nullable(dtype.value_type), nullable=True)
    elif dtype.is_map():
        return dtype.copy(
            key_type=as_nullable(dtype.key_type),
            value_type=as_nullable(dtype.value_type),
            nullable=True,
        )
    else:
        return dtype.copy(nullable=True)


class Backend(SQLBackend, CanCreateCatalog, CanCreateDatabase, CanCreateSchema, NoUrl):
    name = "datafusion"
    supports_arrays = True
    compiler = sc.datafusion.compiler

    @property
    def version(self):
        import importlib.metadata

        return importlib.metadata.version("datafusion")

    def do_connect(
        self, config: Mapping[str, str | Path] | SessionContext | None = None
    ) -> None:
        """Create a DataFusion `Backend` for use with Ibis.

        Parameters
        ----------
        config
            Mapping of table names to files or a `SessionContext`
            instance.

        Examples
        --------
        >>> import ibis
        >>> config = {
        ...     "astronauts": "ci/ibis-testing-data/parquet/astronauts.parquet",
        ...     "diamonds": "ci/ibis-testing-data/csv/diamonds.csv",
        ... }
        >>> con = ibis.datafusion.connect(config)
        >>> con.list_tables()
        ['astronauts', 'diamonds']
        >>> con.table("diamonds")
        DatabaseTable: diamonds
          carat   float64
          cut     string
          color   string
          clarity string
          depth   float64
          table   float64
          price   int64
          x       float64
          y       float64
          z       float64
        """
        if isinstance(config, SessionContext):
            (self.con, config) = (config, None)
        else:
            if config is not None and not isinstance(config, Mapping):
                raise TypeError("Input to ibis.datafusion.connect must be a mapping")
            if SessionConfig is not None:
                df_config = SessionConfig(
                    {"datafusion.sql_parser.dialect": "PostgreSQL"}
                ).with_information_schema(True)
            else:
                df_config = None
            if RuntimeConfig is None:
                self.con = SessionContext(df_config)
            else:
                # datafusion 40.1.0 has a bug where SessionContext requires
                # both SessionConfig and RuntimeConfig be provided.
                self.con = SessionContext(df_config, RuntimeConfig())

        self._register_builtin_udfs()

        if not config:
            config = {}

        for name, path in config.items():
            self._register(path, table_name=name)

    @util.experimental
    @classmethod
    def from_connection(cls, con: SessionContext) -> Backend:
        """Create a DataFusion `Backend` from an existing `SessionContext` instance.

        Parameters
        ----------
        con
            A `SessionContext` instance.
        """
        return ibis.datafusion.connect(con)

    def disconnect(self) -> None:
        pass

    @contextlib.contextmanager
    def _safe_raw_sql(self, sql: sge.Statement) -> Any:
        yield self.raw_sql(sql).collect()

    def _get_schema_using_query(self, query: str) -> sch.Schema:
        name = gen_name("datafusion_metadata_view")
        table = sg.table(name, quoted=self.compiler.quoted)
        src = sge.Create(
            this=table,
            kind="VIEW",
            expression=sg.parse_one(query, read=self.dialect),
            properties=sge.Properties(expressions=[sge.TemporaryProperty()]),
        )

        with self._safe_raw_sql(src):
            pass

        try:
            df = self.con.table(name)
        finally:
            self.drop_view(name)

        return PyArrowSchema.to_ibis(df.schema())

    def _register_builtin_udfs(self):
        from xorq.vendor.ibis.backends.datafusion import udfs

        for name, func in inspect.getmembers(
            udfs,
            predicate=lambda m: (
                callable(m)
                and not m.__name__.startswith("_")
                and m.__module__ == udfs.__name__
            ),
        ):
            annotations = typing.get_type_hints(func)
            argnames = list(inspect.signature(func).parameters.keys())
            input_types = [
                PyArrowType.from_ibis(dt.dtype(annotations.get(arg_name)))
                for arg_name in argnames
            ]
            return_type = PyArrowType.from_ibis(dt.dtype(annotations["return"]))
            udf = df.user_defined.udf(
                func,
                input_types=input_types,
                return_type=return_type,
                volatility="immutable",
                name=name,
            )
            self.con.register_udf(udf)

    def _register_udfs(self, expr: ir.Expr) -> None:
        for udf_node in expr.op().find(ops.ScalarUDF):
            if udf_node.__input_type__ == InputType.PYARROW:
                udf = self._compile_pyarrow_udf(udf_node)
                self.con.register_udf(udf)

        for udf_node in expr.op().find(ops.ElementWiseVectorizedUDF):
            udf = self._compile_elementwise_udf(udf_node)
            self.con.register_udf(udf)

    def _compile_pyarrow_udf(self, udf_node):
        return df.user_defined.udf(
            udf_node.__func__,
            input_types=[PyArrowType.from_ibis(arg.dtype) for arg in udf_node.args],
            return_type=PyArrowType.from_ibis(udf_node.dtype),
            volatility=getattr(udf_node, "__config__", {}).get(
                "volatility", "volatile"
            ),
            name=udf_node.__func_name__,
        )

    def _compile_elementwise_udf(self, udf_node):
        return df.user_defined.udf(
            udf_node.func,
            input_types=list(map(PyArrowType.from_ibis, udf_node.input_type)),
            return_type=PyArrowType.from_ibis(udf_node.return_type),
            volatility="volatile",
            name=udf_node.func.__name__,
        )

    def raw_sql(self, query: str | sge.Expression) -> Any:
        """Execute a SQL string `query` against the database.

        Parameters
        ----------
        query
            Raw SQL string
        kwargs
            Backend specific query arguments

        """
        with contextlib.suppress(AttributeError):
            query = query.sql(dialect=self.dialect, pretty=True)
        self._log(query)
        return self.con.sql(query)

    def list_catalogs(self, like: str | None = None) -> list[str]:
        code = (
            sg.select(C.table_catalog)
            .from_(sg.table("tables", db="information_schema"))
            .distinct()
        ).sql()
        result = self.con.sql(code).to_pydict()
        return self._filter_with_like(result["table_catalog"], like)

    def create_catalog(self, name: str, force: bool = False) -> None:
        with self._safe_raw_sql(
            sge.Create(kind="DATABASE", this=sg.to_identifier(name), exists=force)
        ):
            pass

    def drop_catalog(self, name: str, force: bool = False) -> None:
        raise com.UnsupportedOperationError(
            "DataFusion does not support dropping databases"
        )

    def list_databases(
        self, like: str | None = None, catalog: str | None = None
    ) -> list[str]:
        return self._filter_with_like(
            self.con.catalog(catalog if catalog is not None else "datafusion").names(),
            like=like,
        )

    def create_database(
        self, name: str, catalog: str | None = None, force: bool = False
    ) -> None:
        # not actually a table, but this is how sqlglot represents schema names
        db_name = sg.table(name, db=catalog)
        with self._safe_raw_sql(sge.Create(kind="SCHEMA", this=db_name, exists=force)):
            pass

    def drop_database(
        self, name: str, catalog: str | None = None, force: bool = False
    ) -> None:
        db_name = sg.table(name, db=catalog)
        with self._safe_raw_sql(sge.Drop(kind="SCHEMA", this=db_name, exists=force)):
            pass

    def list_tables(
        self,
        like: str | None = None,
        database: str | None = None,
    ) -> list[str]:
        """Return the list of table names in the current database.

        Parameters
        ----------
        like
            A pattern in Python's regex format.
        database
            Unused in the datafusion backend.

        Returns
        -------
        list[str]
            The list of the table names that match the pattern `like`.
        """
        database = database or "public"
        query = (
            sg.select("table_name")
            .from_("information_schema.tables")
            .where(sg.column("table_schema").eq(sge.convert(database)))
            .order_by("table_name")
        )
        return self._filter_with_like(
            self.raw_sql(query).to_pydict()["table_name"], like
        )

    def get_schema(
        self,
        table_name: str,
        *,
        catalog: str | None = None,
        database: str | None = None,
    ) -> sch.Schema:
        if catalog is not None:
            catalog = self.con.catalog(catalog)
        else:
            catalog = self.con.catalog()

        if database is not None:
            database = catalog.database(database)
        else:
            database = catalog.database()

        table = database.table(table_name)
        return sch.schema(table.schema)

    @deprecated(
        as_of="9.1",
        instead="use the explicit `read_*` method for the filetype you are trying to read, e.g., read_parquet, read_csv, etc.",
    )
    def register(
        self,
        source: str | Path | pa.Table | pa.RecordBatch | pa.Dataset | pd.DataFrame,
        table_name: str | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        return self._register(source, table_name, **kwargs)

    def _register(
        self,
        source: str | Path | pa.Table | pa.RecordBatch | pa.Dataset | pd.DataFrame,
        table_name: str | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        import pandas as pd

        if isinstance(source, (str, Path)):
            first = str(source)
        elif isinstance(source, pa.Table):
            self.con.deregister_table(table_name)
            self.con.register_record_batches(table_name, [source.to_batches()])
            return self.table(table_name)
        elif isinstance(source, pa.RecordBatch):
            self.con.deregister_table(table_name)
            self.con.register_record_batches(table_name, [[source]])
            return self.table(table_name)
        elif isinstance(source, ds.Dataset):
            self.con.deregister_table(table_name)
            self.con.register_dataset(table_name, source)
            return self.table(table_name)
        elif isinstance(source, pd.DataFrame):
            return self.register(pa.Table.from_pandas(source), table_name, **kwargs)
        else:
            raise ValueError("`source` must be either a string or a pathlib.Path")

        if first.startswith(("parquet://", "parq://")) or first.endswith(
            ("parq", "parquet")
        ):
            return self.read_parquet(source, table_name=table_name, **kwargs)
        elif first.startswith(("csv://", "txt://")) or first.endswith(
            ("csv", "tsv", "txt")
        ):
            return self.read_csv(source, table_name=table_name, **kwargs)
        else:
            self._register_failure()
            return None

    def _register_failure(self):
        import inspect

        msg = ", ".join(
            m[0] for m in inspect.getmembers(self) if m[0].startswith("read_")
        )
        raise ValueError(
            f"Cannot infer appropriate read function for input, "
            f"please call one of {msg} directly"
        )

    def _in_memory_table_exists(self, name: str) -> bool:
        db = self.con.catalog().database()
        try:
            db.table(name)
        except Exception:  # noqa: BLE001 because DataFusion has nothing better
            return False
        else:
            return True

    def _register_in_memory_table(self, op: ops.InMemoryTable) -> None:
        # self.con.register_table is broken, so we do this roundabout thing
        # of constructing a datafusion DataFrame, which has a side effect
        # of registering the table
        self.con.from_arrow_table(op.data.to_pyarrow(op.schema), op.name)

    def read_csv(
        self, path: str | Path, table_name: str | None = None, **kwargs: Any
    ) -> ir.Table:
        """Register a CSV file as a table in the current database.

        Parameters
        ----------
        path
            The data source. A string or Path to the CSV file.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to DataFusion loading function.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        path = normalize_filename(path)
        table_name = table_name or gen_name("read_csv")
        # Our other backends support overwriting views / tables when reregistering
        self.con.deregister_table(table_name)
        self.con.register_csv(table_name, path, **kwargs)
        return self.table(table_name)

    def read_parquet(
        self, path: str | Path, table_name: str | None = None, **kwargs: Any
    ) -> ir.Table:
        """Register a parquet file as a table in the current database.

        Parameters
        ----------
        path
            The data source.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to DataFusion loading function.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        path = normalize_filename(path)
        table_name = table_name or gen_name("read_parquet")
        # Our other backends support overwriting views / tables when reregistering
        self.con.deregister_table(table_name)
        self.con.register_parquet(table_name, path, **kwargs)
        return self.table(table_name)

    def read_delta(
        self, source_table: str | Path, table_name: str | None = None, **kwargs: Any
    ) -> ir.Table:
        """Register a Delta Lake table as a table in the current database.

        Parameters
        ----------
        source_table
            The data source. Must be a directory
            containing a Delta Lake table.
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        **kwargs
            Additional keyword arguments passed to deltalake.DeltaTable.

        Returns
        -------
        ir.Table
            The just-registered table

        """
        source_table = normalize_filename(source_table)

        table_name = table_name or gen_name("read_delta")

        # Our other backends support overwriting views / tables when reregistering
        self.con.deregister_table(table_name)

        try:
            from deltalake import DeltaTable
        except ImportError:
            raise ImportError(
                "The deltalake extra is required to use the "
                "read_delta method. You can install it using pip:\n\n"
                "pip install 'ibis-framework[deltalake]'\n"
            )

        delta_table = DeltaTable(source_table, **kwargs)
        self.con.register_dataset(table_name, delta_table.to_pyarrow_dataset())
        return self.table(table_name)

    def to_pyarrow_batches(
        self,
        expr: ir.Expr,
        *,
        chunk_size: int = 1_000_000,
        **kwargs: Any,
    ) -> pa.ipc.RecordBatchReader:
        pa = self._import_pyarrow()

        self._register_udfs(expr)
        self._register_in_memory_tables(expr)

        table_expr = expr.as_table()
        raw_sql = self.compile(table_expr, **kwargs)

        frame = self.con.sql(raw_sql)

        schema = sch.Schema(
            {name: as_nullable(typ) for name, typ in table_expr.schema().items()}
        )
        names = schema.names

        struct_schema = schema.as_struct().to_pyarrow()

        def make_gen():
            yield from (
                # convert the renamed + casted columns into a record batch
                pa.RecordBatch.from_struct_array(
                    # rename columns to match schema because datafusion lowercases things
                    pa.RecordBatch.from_arrays(batch.to_pyarrow().columns, names=names)
                    # cast the struct array to the desired types to work around
                    # https://github.com/apache/arrow-datafusion-python/issues/534
                    .to_struct_array()
                    .cast(struct_schema, safe=False)
                )
                for batch in frame.execute_stream()
            )

        return pa.ipc.RecordBatchReader.from_batches(schema.to_pyarrow(), make_gen())

    def to_pyarrow(self, expr: ir.Expr, **kwargs: Any) -> pa.Table:
        batch_reader = self.to_pyarrow_batches(expr, **kwargs)
        arrow_table = batch_reader.read_all()
        return expr.__pyarrow_result__(arrow_table)

    def execute(self, expr: ir.Expr, **kwargs: Any):
        batch_reader = self.to_pyarrow_batches(expr, **kwargs)
        return expr.__pandas_result__(
            batch_reader.read_pandas(timestamp_as_object=True)
        )

    def create_table(
        self,
        name: str,
        obj: ir.Table
        | pd.DataFrame
        | pa.Table
        | pa.RecordBatchReader
        | pa.RecordBatch
        | pl.DataFrame
        | pl.LazyFrame
        | None = None,
        *,
        schema: sch.SchemaLike | None = None,
        database: str | None = None,
        temp: bool = False,
        overwrite: bool = False,
    ):
        """Create a table in DataFusion.

        Parameters
        ----------
        name
            Name of the table to create
        obj
            The data with which to populate the table; optional, but at least
            one of `obj` or `schema` must be specified
        schema
            The schema of the table to create; optional, but at least one of
            `obj` or `schema` must be specified
        database
            The name of the database in which to create the table; if not
            passed, the current database is used.
        temp
            Create a temporary table
        overwrite
            If `True`, replace the table if it already exists, otherwise fail
            if the table exists

        """
        if obj is None and schema is None:
            raise ValueError("Either `obj` or `schema` must be specified")
        if schema is not None:
            schema = ibis.schema(schema)

        properties = []

        if temp:
            properties.append(sge.TemporaryProperty())

        quoted = self.compiler.quoted

        if isinstance(obj, ir.Expr):
            table = obj

            # If it's a memtable, it will get registered in the pre-execute hooks
            self._run_pre_execute_hooks(table)

            compiler = self.compiler
            relname = "_"
            query = sg.select(
                *(
                    compiler.cast(
                        sg.column(col, table=relname, quoted=quoted), dtype
                    ).as_(col, quoted=quoted)
                    for col, dtype in table.schema().items()
                )
            ).from_(
                compiler.to_sqlglot(table).subquery(
                    sg.to_identifier(relname, quoted=quoted)
                )
            )
        elif obj is not None:
            table_ident = sg.table(name, db=database, quoted=quoted).sql(self.dialect)
            _read_in_memory(obj, table_ident, self, overwrite=overwrite)
            return self.table(name, database=database)
        else:
            query = None

        table_ident = sg.table(name, db=database, quoted=quoted)

        if query is None:
            target = sge.Schema(
                this=table_ident,
                expressions=(schema or table.schema()).to_sqlglot(self.dialect),
            )
        else:
            target = table_ident

        create_stmt = sge.Create(
            kind="TABLE",
            this=target,
            properties=sge.Properties(expressions=properties),
            expression=query,
            replace=overwrite,
        )

        with self._safe_raw_sql(create_stmt):
            pass

        return self.table(name, database=database)

    def truncate_table(
        self, name: str, database: str | None = None, schema: str | None = None
    ) -> None:
        """Delete all rows from a table.

        Parameters
        ----------
        name
            Table name
        database
            Database name
        schema
            Schema name

        """
        # datafusion doesn't support `TRUNCATE TABLE` so we use `DELETE FROM`
        #
        # however datafusion as of 34.0.0 doesn't implement DELETE DML yet
        table_loc = self._warn_and_create_table_loc(database, schema)
        catalog, db = self._to_catalog_db_tuple(table_loc)

        ident = sg.table(name, db=db, catalog=catalog).sql(self.dialect)
        with self._safe_raw_sql(sge.delete(ident)):
            pass


@contextlib.contextmanager
def _create_and_drop_memtable(_conn, table_name, tmp_name, overwrite):
    """Workaround inability to overwrite tables in dataframe API.

    DataFusion has helper methods for loading in-memory data, but these methods
    don't allow overwriting tables.
    The SQL interface allows creating tables from existing tables, so we register
    the data as a table using the dataframe API, then run a

    CREATE [OR REPLACE] TABLE table_name AS SELECT * FROM in_memory_thing

    and that allows us to toggle the overwrite flag.
    """
    src = sge.Create(
        this=table_name,
        kind="TABLE",
        expression=sg.select("*").from_(tmp_name),
        replace=overwrite,
    )

    yield

    _conn.raw_sql(src)
    _conn.drop_table(tmp_name)


@lazy_singledispatch
def _read_in_memory(
    source: Any, table_name: str, _conn: Backend, overwrite: bool = False
):
    raise NotImplementedError("No support for source or imports missing")


@_read_in_memory.register(dict)
def _pydict(source, table_name, _conn, overwrite: bool = False):
    tmp_name = gen_name("pydict")
    with _create_and_drop_memtable(_conn, table_name, tmp_name, overwrite):
        _conn.con.from_pydict(source, name=tmp_name)


@_read_in_memory.register("polars.DataFrame")
def _polars(source, table_name, _conn, overwrite: bool = False):
    tmp_name = gen_name("polars")
    with _create_and_drop_memtable(_conn, table_name, tmp_name, overwrite):
        _conn.con.from_polars(source, name=tmp_name)


@_read_in_memory.register("polars.LazyFrame")
def _polars(source, table_name, _conn, overwrite: bool = False):
    tmp_name = gen_name("polars")
    with _create_and_drop_memtable(_conn, table_name, tmp_name, overwrite):
        _conn.con.from_polars(source.collect(), name=tmp_name)


@_read_in_memory.register("pyarrow.Table")
def _pyarrow_table(source, table_name, _conn, overwrite: bool = False):
    tmp_name = gen_name("pyarrow")
    with _create_and_drop_memtable(_conn, table_name, tmp_name, overwrite):
        _conn.con.from_arrow_table(source, name=tmp_name)


@_read_in_memory.register("pyarrow.RecordBatchReader")
def _pyarrow_rbr(source, table_name, _conn, overwrite: bool = False):
    tmp_name = gen_name("pyarrow")
    with _create_and_drop_memtable(_conn, table_name, tmp_name, overwrite):
        _conn.con.from_arrow_table(source.read_all(), name=tmp_name)


@_read_in_memory.register("pyarrow.RecordBatch")
def _pyarrow_rb(source, table_name, _conn, overwrite: bool = False):
    tmp_name = gen_name("pyarrow")
    with _create_and_drop_memtable(_conn, table_name, tmp_name, overwrite):
        _conn.con.register_record_batches(tmp_name, [[source]])


@_read_in_memory.register("pyarrow.dataset.Dataset")
def _pyarrow_rb(source, table_name, _conn, overwrite: bool = False):
    tmp_name = gen_name("pyarrow")
    with _create_and_drop_memtable(_conn, table_name, tmp_name, overwrite):
        _conn.con.register_dataset(tmp_name, source)


@_read_in_memory.register("pandas.DataFrame")
def _pandas(source: pd.DataFrame, table_name, _conn, overwrite: bool = False):
    tmp_name = gen_name("pandas")
    with _create_and_drop_memtable(_conn, table_name, tmp_name, overwrite):
        _conn.con.from_pandas(source, name=tmp_name)



import ibis
import pytest
from packaging import version
from pytest import param

from xorq.common.utils.ibis_utils import from_ibis


def test_string_concat():
    s1 = ibis.literal("hello")
    s2 = ibis.literal("world")
    expr = s1 + s2

    assert from_ibis(expr) is not None


def test_string_upper_lower():
    s = ibis.literal("Hello")
    upper_expr = s.upper()
    lower_expr = s.lower()

    assert from_ibis(upper_expr) is not None
    assert from_ibis(lower_expr) is not None


def test_string_to_date():
    value = "20170206"
    format_str = "%Y%m%d"
    s = ibis.literal(value)
    expr = s.as_date(format_str)

    assert from_ibis(expr) is not None


def test_string_length():
    s = ibis.literal("hello")
    expr = s.length()

    assert from_ibis(expr) is not None


def test_string_substring():
    s = ibis.literal("hello world")
    expr = s.substr(0, 5)

    assert from_ibis(expr) is not None


def test_string_startswith():
    t = ibis.table({"id": "int32", "name": "string"}, name="T")
    q = t.filter(t.name.startswith("X"))

    assert from_ibis(q) is not None


def test_string_endswith():
    t = ibis.table({"id": "int32", "name": "string"}, name="T")
    q = t.filter(t.name.endswith("X"))

    assert from_ibis(q) is not None


def cases(t):
    args = [(1, "abcd"), (2, "ABCD")]
    kwargs = {"else_": "dabc"}
    if version.parse(ibis.__version__) <= version.parse("9.5.0"):
        args = [[(1, "abcd"), (2, "ABCD")]]
        kwargs = {"default": "dabc"}

    return t.int_col.cases(*args, **kwargs).startswith("abc")


@pytest.mark.parametrize(
    "fun",
    [
        param(
            lambda t: t.string_col.contains("6"),
            id="contains",
        ),
        param(
            lambda t: t.string_col.rlike("|".join(map(str, range(10)))),
            id="rlike",
        ),
        param(
            lambda t: ("a" + t.string_col + "a").re_search(r"\d+"),
            id="re_search_substring",
        ),
        param(
            lambda t: t.string_col.re_search(r"\d+"),
            id="re_search",
        ),
        param(
            lambda t: t.string_col.re_search(r"[[:digit:]]+"),
            id="re_search_posix",
        ),
        param(
            lambda t: t.string_col.re_extract(r"([[:digit:]]+)", 1),
            id="re_extract_posix",
        ),
        param(
            lambda t: (t.string_col + "1").re_extract(r"\d(\d+)", 0),
            id="re_extract_whole_group",
        ),
        param(
            lambda t: t.date_string_col.re_extract(r"(\d+)\D(\d+)\D(\d+)", 1),
            id="re_extract_group_1",
        ),
        param(
            lambda t: t.date_string_col.re_extract(r"(\d+)\D(\d+)\D(\d+)", 2),
            id="re_extract_group_2",
        ),
        param(
            lambda t: t.date_string_col.re_extract(r"(\d+)\D(\d+)\D(\d+)", 3),
            id="re_extract_group_3",
        ),
        param(
            lambda t: t.date_string_col.re_extract(r"^(\d+)", 1),
            id="re_extract_group_at_beginning",
        ),
        param(
            lambda t: t.date_string_col.re_extract(r"(\d+)$", 1),
            id="re_extract_group_at_end",
        ),
        param(
            lambda t: t.string_col.re_replace(r"[[:digit:]]+", "a"),
            id="re_replace_posix",
        ),
        param(
            lambda t: t.string_col.re_replace(r"\d+", "a"),
            id="re_replace",
        ),
        param(
            lambda t: t.string_col.repeat(2),
            id="repeat_method",
        ),
        param(
            lambda t: 2 * t.string_col,
            id="repeat_left",
        ),
        param(
            lambda t: t.string_col * 2,
            id="repeat_right",
        ),
        param(
            lambda t: t.string_col.translate("01", "ab"),
            id="translate",
        ),
        param(
            lambda t: t.string_col.find("a"),
            id="find",
        ),
        param(
            lambda t: t.date_string_col.find("13", 3),
            id="find_start",
        ),
        param(
            lambda t: t.string_col.lpad(10, "a"),
            id="lpad",
        ),
        param(
            lambda t: t.string_col.rpad(10, "a"),
            id="rpad",
        ),
        param(
            lambda t: t.string_col.lower(),
            id="lower",
        ),
        param(
            lambda t: t.string_col.upper(),
            id="upper",
        ),
        param(
            lambda t: t.string_col.reverse(),
            id="reverse",
        ),
        param(
            lambda t: t.string_col.ascii_str(),
            id="ascii_str",
        ),
        param(
            lambda t: t.string_col.length(),
            id="length",
        ),
        param(
            cases,
            id="startswith",
        ),
        param(
            lambda t: t.date_string_col.startswith("2010-01"),
            id="startswith-simple",
        ),
        param(
            lambda t: t.string_col.strip(),
            id="strip",
        ),
        param(
            lambda t: t.string_col.lstrip(),
            id="lstrip",
        ),
        param(
            lambda t: t.string_col.rstrip(),
            id="rstrip",
        ),
        param(
            lambda t: t.string_col.capitalize(),
            id="capitalize",
        ),
        param(
            lambda t: t.date_string_col.substr(2, 3),
            id="substr",
        ),
        param(
            lambda t: t.date_string_col.substr(2),
            id="substr-start-only",
        ),
        param(
            lambda t: t.date_string_col.left(2),
            id="left",
        ),
        param(
            lambda t: t.date_string_col.right(2),
            id="right",
        ),
        param(
            lambda t: t.date_string_col[1:3],
            id="slice",
        ),
        param(
            lambda t: t.date_string_col[-2],
            id="negative-index",
        ),
        param(
            lambda t: t.date_string_col[: t.date_string_col.length()],
            id="expr_slice_end",
        ),
        param(
            lambda t: t.date_string_col[:],
            id="expr_empty_slice",
        ),
        param(
            lambda t: t.date_string_col[
                t.date_string_col.length() - 2 : t.date_string_col.length() - 1
            ],
            id="expr_slice_begin_end",
        ),
        param(
            lambda t: ibis.literal("-").join(["a", t.string_col, "c"]),
            id="join",
        ),
        param(
            lambda t: t.string_col + t.date_string_col,
            id="concat_columns",
        ),
        param(
            lambda t: t.string_col + "a",
            id="concat_column_scalar",
        ),
        param(
            lambda t: "a" + t.string_col,
            id="concat_scalar_column",
        ),
        param(
            lambda t: t.string_col.replace("1", "42"),
            id="replace",
        ),
    ],
)
def test_string(ibis_alltypes, fun):
    expr = fun(ibis_alltypes).name("tmp")

    assert from_ibis(expr).execute() is not None
